from aries_cloudagent.messaging.models.base import BaseModel, BaseModelSchema
from marshmallow import fields, EXCLUDE, validate


class DataAgreementContextDecorator(BaseModel):
    """
    Class representing a data agreement context decorator.
    """

    class Meta:
        """DataAgreementContextDecorator metadata."""

        schema_class = "DataAgreementContextDecoratorSchema"

    def __init__(
        self,
        *,
        message_type: str = None,
        message: dict = None,
        **kwargs
    ):
        """
        Initialize a DataAgreementContextDecorator instance.

        Args:
            message_type: The type of the message.
            message: The message.
            kwargs: The keyword arguments

        """
        super().__init__(**kwargs)
        self.message_type = message_type
        self.message = message


class DataAgreementContextDecoratorSchema(BaseModelSchema):
    """
    DataAgreementContextDecorator schema.
    """

    class Meta:
        """DataAgreementContextDecoratorSchema metadata."""

        model_class = DataAgreementContextDecorator
        unknown = EXCLUDE

    message_type = fields.Str(
        required=False,
        description="The type of the message.",
        example="protocol",
        validate=validate.OneOf(["protocol", "non-protocol"]),
    )
    message = fields.Dict(
        required=False,
        description="The message.",
        example={"data_agreement_id": "12345678-1234-1234-1234-123456789012"},
    )



import logging
from aiohttp import web
from aiohttp_apispec import (
    docs,
    request_schema,
    match_info_schema,
)

from ..manager import ADAManager, ADAManagerError
from ..routes.maps.tag_maps import (
    TAGS_JSONLD_FUNCTIONS_LABEL,
)

from .openapi import (
    SendJSONLDDIDCommProcessedDataMessageHandlerRequestSchema,
    SendJSONLDDIDCommProcessedDataMessageHandlerMatchInfoSchema,
)

LOGGER = logging.getLogger(__name__)

PAGINATION_PAGE_SIZE = 10


@docs(
    tags=[TAGS_JSONLD_FUNCTIONS_LABEL],
    summary="Send JSON-LD processed-data didcomm message to the remote agent.",
    responses={
        204: {
            "description": "Success",
        },
    },
)
@match_info_schema(SendJSONLDDIDCommProcessedDataMessageHandlerMatchInfoSchema())
@request_schema(SendJSONLDDIDCommProcessedDataMessageHandlerRequestSchema())
async def send_json_ld_didcomm_processed_data_message_handler(request: web.BaseRequest):
    """Send JSON-LD didcomm processed data message handler."""

    # Context.
    context = request.app["request_context"]

    # Fetch path parameters.
    connection_id = request.match_info["connection_id"]

    # Fetch request body
    body = await request.json()

    # Initialise MyData DID Manager.
    mydata_did_manager: ADAManager = ADAManager(context=context)

    try:

        # Call the function.

        await mydata_did_manager.send_json_ld_processed_message(
            connection_id=connection_id,
            data=body.get("data", {}),
            signature_options=body.get("signature_options", {}),
            proof_chain=body.get("proof_chain", False),
        )

    except ADAManagerError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response({}, status=204)



"""Key type enum."""

from enum import Enum
from typing import NamedTuple, Optional

# Define keys
KeySpec = NamedTuple(
    "KeySpec",
    [("key_type", str), ("multicodec_name", str), ("multicodec_prefix", int)],
)


class KeyTypeException(BaseException):
    """Key type exception."""


class KeyType(Enum):
    """KeyType Enum specifying key types with multicodec name."""

    # NOTE: the py_multicodec library is outdated. We use hardcoded prefixes here
    # until this PR gets released: https://github.com/multiformats/py-multicodec/pull/14
    # multicodec is also not used now, but may be used again if py_multicodec is updated
    ED25519 = KeySpec("ed25519", "ed25519-pub", b"\xed\x01")

    @property
    def key_type(self) -> str:
        """Getter for key type identifier."""
        return self.value.key_type

    @property
    def multicodec_name(self) -> str:
        """Getter for multicodec name."""
        return self.value.multicodec_name

    @property
    def multicodec_prefix(self) -> bytes:
        """Getter for multicodec prefix."""
        return self.value.multicodec_prefix

    @classmethod
    def from_multicodec_name(cls, multicodec_name: str) -> Optional["KeyType"]:
        """Get KeyType instance based on multicodec name. Returns None if not found."""
        for key_type in KeyType:
            if key_type.multicodec_name == multicodec_name:
                return key_type

        return None

    @classmethod
    def from_multicodec_prefix(cls, multicodec_prefix: bytes) -> Optional["KeyType"]:
        """Get KeyType instance based on multicodec prefix. Returns None if not found."""
        for key_type in KeyType:
            if key_type.multicodec_prefix == multicodec_prefix:
                return key_type

        return None

    @classmethod
    def from_prefixed_bytes(cls, prefixed_bytes: bytes) -> Optional["KeyType"]:
        """Get KeyType instance based on prefix in bytes. Returns None if not found."""
        for key_type in KeyType:
            if prefixed_bytes.startswith(key_type.multicodec_prefix):
                return key_type

        return None

    @classmethod
    def from_key_type(cls, key_type: str) -> Optional["KeyType"]:
        """Get KeyType instance from the key type identifier."""
        for _key_type in KeyType:
            if _key_type.key_type == key_type:
                return _key_type

        return None



import logging
from aiohttp import web
from aiohttp_apispec import (
    docs,
    request_schema,
    querystring_schema,
    response_schema,
    match_info_schema,
)
from aries_cloudagent.messaging.models.base import BaseModelError
from aries_cloudagent.storage.error import StorageError
from dexa_sdk.managers.ada_manager import V2ADAManager
from dexa_sdk.managers.dexa_manager import DexaManager
from dexa_sdk.utils import clean_and_get_field_from_dict
from ..manager import ADAManagerError
from ..models.exchange_records.data_agreement_personal_data_record import (
    DataAgreementPersonalDataRecordSchema,
)
from ..utils.util import (
    str_to_bool,
)
from ..routes.maps.tag_maps import (
    TAGS_DATA_AGREEMENT_CORE_FUNCTIONS_LABEL,
)

from .openapi import (
    DataAgreementV1RecordResponseSchema,
    DataAgreementQueryStringSchema,
    UpdateDataAgreementMatchInfoSchema,
    DeleteDataAgreementMatchInfoSchema,
    DataAgreementQRCodeMatchInfoSchema,
    GenerateDataAgreementQrCodePayloadResponseSchema,
    CreateOrUpdateDataAgreementInWalletQueryStringSchema,
    PublishDataAgreementMatchInfoSchema,
    QueryDaPersonalDataInWalletQueryStringSchema,
    UpdateDaPersonalDataInWalletMatchInfoSchema,
    UpdateDaPersonalDataInWalletRequestSchema,
    UpdateDaPersonalDataInWalletResponseSchema,
    DeleteDaPersonalDataInWalletMatchInfoSchema,
    QueryDataAgreementQrCodeMetadataRecordsMatchInfoSchema,
    QueryDataAgreementQRCodeMetadataRecordsResponseSchema,
    RemoveDataAgreementQrCodeMetadataRecordMatchInfoSchema,
    SendDataAgreementQrCodeWorkflowInitiateHandlerMatchInfoSchema,
    GenerateDataAgreementQrCodePayloadQueryStringSchema,
    CreateOrUpdateDataAgreementInWalletRequestSchema,
    UpdateDataAgreementTemplateOpenAPISchema,
    ConfigureCustomerIdentificationDAMatchInfoSchema
)

LOGGER = logging.getLogger(__name__)

PAGINATION_PAGE_SIZE = 10


@docs(
    tags=[TAGS_DATA_AGREEMENT_CORE_FUNCTIONS_LABEL],
    summary="Create and store data agreement template in wallet",
    responses={422: {"description": "Unprocessable Entity (invalid request payload)"}},
)
@querystring_schema(CreateOrUpdateDataAgreementInWalletQueryStringSchema())
@request_schema(CreateOrUpdateDataAgreementInWalletRequestSchema())
@response_schema(DataAgreementV1RecordResponseSchema(), 201)
async def create_and_store_data_agreement_in_wallet_v2(request: web.BaseRequest):
    """Create and store data agreement template in wallet."""

    # Request context
    context = request.app["request_context"]

    # Fetch request body
    data_agreement = await request.json()

    # Initialise MyData DID Manager
    manager: V2ADAManager = V2ADAManager(context=context)

    # Fetch querystring params
    publish_flag = str_to_bool(clean_and_get_field_from_dict(request.query, "publish_flag"))
    existing_schema_id = clean_and_get_field_from_dict(request.query, "existing_schema_id")

    try:

        # Create and store data agreement in wallet
        da_record = await manager.create_and_store_da_template_in_wallet(
            data_agreement=data_agreement,
            publish_flag=publish_flag,
            schema_id=existing_schema_id,
        )

        if not da_record:
            raise web.HTTPBadRequest(reason="Data agreement not created")

    except ADAManagerError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response(da_record.serialize(), status=201)


@docs(
    tags=[TAGS_DATA_AGREEMENT_CORE_FUNCTIONS_LABEL],
    summary="Publish data agreement templates in the wallet",
    responses={400: {"description": "Bad Request (invalid request payload)"}},
)
@match_info_schema(PublishDataAgreementMatchInfoSchema())
@response_schema(DataAgreementV1RecordResponseSchema(), 200)
async def publish_data_agreement_handler(request: web.BaseRequest):
    """Publish data agreement template in the wallet."""

    # Request context
    context = request.app["request_context"]

    # Get path parameters
    template_id = request.match_info["template_id"]

    # Initialise MyData DID Manager
    manager: V2ADAManager = V2ADAManager(context=context)

    try:

        # Publish data agreement in the wallet
        record = await manager.publish_da_template_in_wallet(template_id)

        if not record:
            raise web.HTTPBadRequest(reason="Data agreement not published")

    except ADAManagerError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response(record.serialize(), status=200)


@docs(
    tags=[TAGS_DATA_AGREEMENT_CORE_FUNCTIONS_LABEL],
    summary="Query data agreement templates in the wallet",
)
@querystring_schema(DataAgreementQueryStringSchema())
@response_schema(DataAgreementV1RecordResponseSchema(many=True), 200)
async def query_data_agreements_in_wallet(request: web.BaseRequest):
    """Query data agreement templates in the wallet."""

    # Request context
    context = request.app["request_context"]

    # Fetch query string parameters
    method_of_use = clean_and_get_field_from_dict(request.query, "method_of_use")
    template_id = clean_and_get_field_from_dict(request.query, "template_id")
    template_version = clean_and_get_field_from_dict(request.query, "template_version")
    delete_flag = clean_and_get_field_from_dict(request.query, "delete_flag")
    publish_flag = clean_and_get_field_from_dict(request.query, "publish_flag")
    latest_version_flag = clean_and_get_field_from_dict(request.query, "latest_version_flag")
    third_party_data_sharing = clean_and_get_field_from_dict(
        request.query, "third_party_data_sharing")
    page = clean_and_get_field_from_dict(request.query, "page")
    page = int(page) if page is not None else page
    page_size = clean_and_get_field_from_dict(request.query, "page_size")
    page_size = int(page_size) if page_size is not None else page_size

    # Initialise MyData DID Manager
    manager: V2ADAManager = V2ADAManager(context=context)

    try:

        # Query data agreements in the wallet
        paginationResult = await manager.query_da_templates_in_wallet(
            template_id=template_id,
            delete_flag=delete_flag,
            method_of_use=method_of_use,
            publish_flag=publish_flag,
            latest_version_flag=latest_version_flag,
            template_version=template_version,
            third_party_data_sharing=third_party_data_sharing,
            page=page if page else 1,
            page_size=page_size if page_size else 10
        )

    except (StorageError, BaseModelError, ValueError) as err:

        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response(paginationResult._asdict())


@docs(
    tags=[TAGS_DATA_AGREEMENT_CORE_FUNCTIONS_LABEL],
    summary="Update data agreement template in the wallet",
    responses={400: {"description": "Bad Request (invalid request payload)"}},
)
@match_info_schema(UpdateDataAgreementMatchInfoSchema())
@querystring_schema(UpdateDataAgreementTemplateOpenAPISchema())
@request_schema(CreateOrUpdateDataAgreementInWalletRequestSchema())
@response_schema(DataAgreementV1RecordResponseSchema(), 200)
async def update_data_agreement_in_wallet_v2(request: web.BaseRequest):
    """Update data agreement template in the wallet."""

    # Request context
    context = request.app["request_context"]

    # URL params
    template_id = request.match_info["template_id"]

    # Fetch request body
    data_agreement = await request.json()

    # Initialise MyData DID Manager
    manager: V2ADAManager = V2ADAManager(context=context)

    # Fetch querystring params
    publish_flag = str_to_bool(clean_and_get_field_from_dict(request.query, "publish_flag"))
    existing_schema_id = clean_and_get_field_from_dict(request.query, "existing_schema_id")

    try:
        # Update data agreement in the wallet
        record = await manager.update_and_store_da_template_in_wallet(
            template_id=template_id,
            data_agreement=data_agreement,
            publish_flag=publish_flag,
            schema_id=existing_schema_id
        )

    except ADAManagerError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response(record.serialize(), status=200)


@docs(
    tags=[TAGS_DATA_AGREEMENT_CORE_FUNCTIONS_LABEL],
    summary="Delete data agreement template in the wallet",
    responses={
        204: {"description": "No Content (data agreement deleted)"},
        400: {"description": "Bad Request (invalid request payload)"},
    },
)
@match_info_schema(DeleteDataAgreementMatchInfoSchema())
async def delete_data_agreement_in_wallet(request: web.BaseRequest):
    """Delete data agreement template in the wallet."""

    # Request context
    context = request.app["request_context"]

    # URL params
    template_id = request.match_info["template_id"]

    # Initialise MyData DID Manager
    manager: V2ADAManager = V2ADAManager(context=context)

    try:
        # Delete data agreement template in the wallet
        await manager.delete_da_template_in_wallet(
            template_id=template_id
        )

    except ADAManagerError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response({}, status=204)


@docs(
    tags=[TAGS_DATA_AGREEMENT_CORE_FUNCTIONS_LABEL],
    summary="Query data agreement personal data in wallet",
)
@querystring_schema(QueryDaPersonalDataInWalletQueryStringSchema())
@response_schema(DataAgreementPersonalDataRecordSchema(many=True), 200)
async def query_da_personal_data_in_wallet(request: web.BaseRequest):
    """Query data agreement personal data in wallet."""

    # Request context
    context = request.app["request_context"]

    # Initialise MyData DID Manager
    manager: V2ADAManager = V2ADAManager(context=context)

    # Fetch query string parameters
    method_of_use = clean_and_get_field_from_dict(request.query, "method_of_use")
    template_id = clean_and_get_field_from_dict(request.query, "template_id")
    third_party_data_sharing = clean_and_get_field_from_dict(
        request.query, "third_party_data_sharing")
    page = clean_and_get_field_from_dict(request.query, "page")
    page = int(page) if page is not None else page
    page_size = clean_and_get_field_from_dict(request.query, "page_size")
    page_size = int(page_size) if page_size is not None else page_size

    try:

        # Query data agreement personal data in wallet
        paginationResult = await manager.query_pd_of_da_template_from_wallet(
            template_id=template_id,
            method_of_use=method_of_use,
            third_party_data_sharing=third_party_data_sharing,
            page=page if page else 1,
            page_size=page_size if page_size else 10
        )

    except ADAManagerError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response(paginationResult._asdict())


@docs(
    tags=[TAGS_DATA_AGREEMENT_CORE_FUNCTIONS_LABEL],
    summary="Update personal data description in wallet.",
    responses={400: {"description": "Bad Request (invalid request payload)"}},
)
@match_info_schema(UpdateDaPersonalDataInWalletMatchInfoSchema())
@request_schema(UpdateDaPersonalDataInWalletRequestSchema())
@response_schema(UpdateDaPersonalDataInWalletResponseSchema(), 200)
async def update_da_personal_data_in_wallet(request: web.BaseRequest):
    """Update personal data description."""
    # Request context
    context = request.app["request_context"]

    # Initialise MyData DID Manager
    manager: V2ADAManager = V2ADAManager(context=context)

    # URL params
    attribute_id = request.match_info["attribute_id"]

    # Request data
    body = await request.json()

    attribute_description = body.get("attribute_description")

    try:
        # Update data agreement personal data in wallet
        record = await manager.update_personal_data_description(
            attribute_id=attribute_id,
            desc=attribute_description
        )
    except ADAManagerError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response(record.serialize())


@docs(
    tags=[TAGS_DATA_AGREEMENT_CORE_FUNCTIONS_LABEL],
    summary="Delete personal data in wallet",
    responses={
        204: {"description": "No Content (data agreement personal data deleted)"},
        400: {"description": "Bad Request (invalid request payload)"},
    },
)
@match_info_schema(DeleteDaPersonalDataInWalletMatchInfoSchema())
async def delete_da_personal_data_in_wallet(request: web.BaseRequest):
    """Delete personal data in wallet."""

    # Request context
    context = request.app["request_context"]

    # URL params
    attribute_id = request.match_info["attribute_id"]

    # Initialise MyData DID Manager
    manager: V2ADAManager = V2ADAManager(context=context)

    try:

        await manager.delete_personal_data(
            attribute_id=attribute_id
        )

    except ADAManagerError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response({}, status=204)


@docs(
    tags=[TAGS_DATA_AGREEMENT_CORE_FUNCTIONS_LABEL],
    summary="Generate Data Agreement QR code payload",
    responses={400: "Bad Request"},
)
@querystring_schema(GenerateDataAgreementQrCodePayloadQueryStringSchema())
@match_info_schema(DataAgreementQRCodeMatchInfoSchema())
@response_schema(GenerateDataAgreementQrCodePayloadResponseSchema(), 201)
async def generate_data_agreement_qr_code_payload(request: web.BaseRequest):
    # Get path parameters.
    template_id = request.match_info["template_id"]

    # Context.
    context = request.app["request_context"]

    # Query string params
    multi_use_flag = clean_and_get_field_from_dict(request.query, "multi_use")
    multi_use_flag = str_to_bool(multi_use_flag)

    # Initialise MyData DID Manager.
    manager = V2ADAManager(context)

    try:

        # Call the function.
        result = await manager.create_data_agreement_qr_code(
            template_id,
            multi_use_flag
        )

    except ADAManagerError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response(result, status=201)


@docs(
    tags=[TAGS_DATA_AGREEMENT_CORE_FUNCTIONS_LABEL],
    summary="Query Data Agreement QR code metadata records",
)
@match_info_schema(QueryDataAgreementQrCodeMetadataRecordsMatchInfoSchema())
@response_schema(QueryDataAgreementQRCodeMetadataRecordsResponseSchema(many=True))
async def query_data_agreement_qr_code_metadata_records_handler(
    request: web.BaseRequest,
):
    # Context.
    context = request.app["request_context"]

    # Path parameters
    template_id = request.match_info["template_id"]

    # Initialise MyData DID Manager.
    mgr = V2ADAManager(context=context)

    try:
        # Call the function.
        result = await mgr.query_data_agreement_qr_codes(template_id)

    except ADAManagerError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response(result._asdict())


@docs(
    tags=[TAGS_DATA_AGREEMENT_CORE_FUNCTIONS_LABEL],
    summary="Delete Data Agreement QR code record.",
    responses={
        204: {"description": "Success"},
    },
)
@match_info_schema(RemoveDataAgreementQrCodeMetadataRecordMatchInfoSchema())
async def remove_data_agreement_qr_code_metadata_record_handler(
    request: web.BaseRequest,
):

    # Context
    context = request.app["request_context"]

    # Fetch path parameters.
    template_id = request.match_info["template_id"]
    qr_id = request.match_info["qr_id"]

    # Initialise MyData DID Manager.
    mgr = V2ADAManager(context=context)

    try:
        # Call the function.
        await mgr.delete_data_agreement_qr_code(template_id, qr_id)

    except ADAManagerError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response({}, status=204)


@docs(
    tags=[TAGS_DATA_AGREEMENT_CORE_FUNCTIONS_LABEL],
    summary="Send data agreement qr code workflow initiate message to remote agent",
)
@match_info_schema(SendDataAgreementQrCodeWorkflowInitiateHandlerMatchInfoSchema())
async def send_data_agreements_qr_code_workflow_initiate_handler(
    request: web.BaseRequest,
):

    # Context.
    context = request.app["request_context"]

    # Fetch path parameters.

    connection_id = request.match_info["connection_id"]
    qr_id = request.match_info["qr_id"]

    # Initialise MyData DID Manager.
    mgr = V2ADAManager(context=context)

    try:

        # Call the function.
        await mgr.send_qr_code_initiate_message(
            qr_id,
            connection_id
        )

    except ADAManagerError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response({}, status=204)


@docs(
    tags=[TAGS_DATA_AGREEMENT_CORE_FUNCTIONS_LABEL],
    summary="Fetch customer identification data agreement.",
)
async def fetch_customer_identification_da_handler(request: web.BaseRequest):
    """Fetch customer identification DA.

    Args:
        request (web.BaseRequest): Request.
    """

    # Context
    context = request.app["request_context"]

    # Initialise the manager
    mgr = DexaManager(context)

    # Call the function
    record = await mgr.fetch_customer_identification_data_agreement()

    return web.json_response(record.serialize() if record else {})


@docs(
    tags=[TAGS_DATA_AGREEMENT_CORE_FUNCTIONS_LABEL],
    summary="Configure customer identification data agreement.",
)
@match_info_schema(ConfigureCustomerIdentificationDAMatchInfoSchema())
async def configure_customer_identification_da_handler(request: web.BaseRequest):
    """Configure customer identification DA.

    Args:
        request (web.BaseRequest): Request.
    """

    # Context
    context = request.app["request_context"]

    # Path parameters
    template_id = request.match_info["template_id"]

    # Initialise the manager
    mgr = DexaManager(context)

    # Call the function
    record = await mgr.configure_customer_identification_data_agreement(template_id)

    return web.json_response(record.serialize())



from os import environ
from typing import Any
import typing

from marshmallow import fields, validate

from aries_cloudagent.messaging.models.base_record import BaseExchangeRecord, BaseExchangeSchema
from aries_cloudagent.messaging.valid import UUIDFour


class AuditorDIDCommTransactionRecord(BaseExchangeRecord):
    """
    AuditorDIDCommTransactionRecord

    Transaction record to keep track of audit operations against an Auditor.
    """

    class Meta:
        # Schema class
        schema_class = "AuditorDIDCommTransactionRecordSchema"

    # Record type
    RECORD_TYPE = "auditor_didcomm_transaction_record"

    # Wallet record identifier field
    RECORD_ID_NAME = "auditor_didcomm_transaction_record_id"

    # Webhook topic name for this record type
    WEBHOOK_TOPIC = None

    # Wallet record tags used for filtering
    TAG_NAMES = {"~thread_id", "~connection_id"}

    def __init__(
        self,
        *,
        auditor_didcomm_transaction_record_id: str = None,
        thread_id: str = None,
        messages_list: typing.List[dict] = None,
        connection_id: str = None,
        **kwargs
    ):
        super().__init__(auditor_didcomm_transaction_record_id, **kwargs)

        self.thread_id = thread_id
        self.messages_list = messages_list
        self.connection_id = connection_id

    @property
    def auditor_didcomm_transaction_record_id(self) -> str:
        """
        Returns transaction record identifier.
        """
        return self._id

    @property
    def record_value(self) -> dict:
        """Accessor for JSON record value generated for this transaction record."""
        return {
            prop: getattr(self, prop)
            for prop in (
                "thread_id",
                "messages_list",
                "connection_id",
            )
        }

    def __eq__(self, other: Any) -> bool:
        """Comparison between records."""
        return super().__eq__(other)


class AuditorDIDCommTransactionRecordSchema(BaseExchangeSchema):
    """
    AuditorDIDCommTransactionRecordSchema

    Schema class for the AuditorDIDCommTransactionRecord class.
    """

    class Meta:
        # Model class
        model_class = AuditorDIDCommTransactionRecord

    # Wallet record identifier field
    auditor_didcomm_transaction_record_id = fields.Str(
        required=True,
        description="Transaction record id",
        example=UUIDFour.EXAMPLE,
    )

    # Parent message id of the transaction record
    thread_id = fields.Str(
        required=False,
        description="Parent message identifier",
        example=UUIDFour.EXAMPLE,
    )

    # List of messages for the transaction record
    messages_list = fields.List(
        fields.Dict(), required=False, description="Messages list")

    # Connection identifier
    connection_id = fields.Str(
        required=False,
        description="Connection identifier",
        example=UUIDFour.EXAMPLE,
    )



"""Presentation message handler."""
from aries_cloudagent.messaging.base_handler import (
    BaseHandler,
    BaseResponder,
    RequestContext,
)
from aries_cloudagent.utils.tracing import trace_event, get_timer
from mydata_did.patched_protocols.present_proof.v1_0.models.presentation_exchange import (
    V10PresentationExchange
)
from dexa_sdk.managers.ada_manager import V2ADAManager
from ..manager import PresentationManager
from ..messages.presentation import Presentation


class PresentationHandler(BaseHandler):
    """Message handler class for presentations."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """
        Message handler logic for presentations.

        Args:
            context: request context
            responder: responder callback

        """
        r_time = get_timer()

        self._logger.debug("PresentationHandler called with context %s", context)
        assert isinstance(context.message, Presentation)
        self._logger.info(
            "Received presentation message: %s",
            context.message.serialize(as_string=True),
        )

        presentation_manager = PresentationManager(context)

        presentation_exchange_record: V10PresentationExchange = \
            await presentation_manager.receive_presentation()

        # Initialise ADA manager
        manager = V2ADAManager(context)

        # Process the decorator for data agreement accept message if available.
        await manager.process_decorator_with_da_accept_message(
            context.message._decorators,
            presentation_exchange_record,
            context.connection_record
        )

        r_time = trace_event(
            context.settings,
            context.message,
            outcome="PresentationHandler.handle.END",
            perf_counter=r_time,
        )

        if context.settings.get("debug.auto_verify_presentation"):
            await presentation_manager.verify_presentation(presentation_exchange_record)

            trace_event(
                context.settings,
                presentation_exchange_record,
                outcome="PresentationHandler.handle.VERIFY",
                perf_counter=r_time,
            )



from aries_cloudagent.config.injection_context import InjectionContext
from aries_cloudagent.config.provider import ClassProvider
from aries_cloudagent.core.protocol_registry import ProtocolRegistry
from aries_cloudagent.core.plugin_registry import PluginRegistry

from .v1_0.message_types import MESSAGE_TYPES
from .definition import versions
from .patched_protocols.issue_credential.v1_0.message_types import (
    MESSAGE_TYPES as ISSUE_CREDENTIAL_MESSAGE_TYPES,
)
from .patched_protocols.present_proof.v1_0.message_types import (
    MESSAGE_TYPES as PRESENT_PROOF_MESSAGE_TYPES,
)


async def setup(context: InjectionContext):
    # Register patched message types.
    protocol_registry: ProtocolRegistry = await context.inject(ProtocolRegistry)
    protocol_registry.register_message_types(
        MESSAGE_TYPES, version_definition=versions[0]
    )
    protocol_registry.register_message_types(
        ISSUE_CREDENTIAL_MESSAGE_TYPES, version_definition=versions[0]
    )
    protocol_registry.register_message_types(
        PRESENT_PROOF_MESSAGE_TYPES, version_definition=versions[0]
    )

    # Register patched protocol plugins
    plugin_registry: PluginRegistry = await context.inject(PluginRegistry)
    plugin_registry.register_plugin(
        "mydata_did.patched_protocols.issue_credential.v1_0"
    )
    plugin_registry.register_plugin("mydata_did.patched_protocols.present_proof.v1_0")

    # Unregister superseded protocols
    plugin_registry._plugins.pop("aries_cloudagent.protocols.issue_credential")
    plugin_registry._plugins.pop("aries_cloudagent.protocols.present_proof")



from aries_cloudagent.messaging.base_handler import (
    BaseHandler,
    BaseResponder,
    RequestContext,
)

from ..messages.problem_report import MyDataDIDProblemReportMessage
from ..manager import ADAManager


class MyDataDIDProblemReportHandler(BaseHandler):
    """
    Handles problem report messages for MyData DID protocol.
    """

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """
        Message handler logic
        """

        # Assert if received message is of type MyDataDIDProblemReportMessage
        assert isinstance(context.message, MyDataDIDProblemReportMessage)

        self._logger.info(
            "Received problem-report message: %s",
            context.message.serialize(as_string=True)
        )

        # Check if connection is ready
        if not context.connection_ready:
            self._logger.info(
                "Connection not active, skipping problem-report handler: %s",
                context.message_receipt.sender_did,
            )
            return
        
        # Initialize ADA manager
        mgr = ADAManager(context)

        # Process the problem report message for MyData DID protocol
        await mgr.process_mydata_did_problem_report_message(context.message, context.message_receipt)



import logging
from aiohttp import web
from aiohttp_apispec import (
    docs,
    request_schema
)
from dexa_sdk.managers.ada_manager import V2ADAManager
from ..routes.maps.tag_maps import (
    TAGS_DATA_CONTROLLER_FUNCTIONS_LABEL,
)
from .openapi.schemas import (
    UpdateControllerDetailsRequestSchema
)


LOGGER = logging.getLogger(__name__)

PAGINATION_PAGE_SIZE = 10


@docs(
    tags=[TAGS_DATA_CONTROLLER_FUNCTIONS_LABEL],
    summary="Send data controller details message to remote agent hosted by Data Controller",
    responses={
        200: {
            "description": "Success",
        }
    },
)
async def send_data_controller_details_message_handler(request: web.BaseRequest):
    """Send data controller details message to remote agent hosted by Data Controller."""

    context = request.app["request_context"]
    connection_id = request.match_info["connection_id"]

    # Initialise MyData DID Manager.
    mgr = V2ADAManager(context=context)

    # Call the function
    await mgr.send_data_controller_details_message(connection_id)

    return web.json_response({}, status=200)


@docs(
    tags=[TAGS_DATA_CONTROLLER_FUNCTIONS_LABEL],
    summary="Update data controller details",
)
@request_schema(UpdateControllerDetailsRequestSchema())
async def update_data_controller_details(request: web.BaseRequest):
    """Update data controller details"""
    # Request context
    context = request.app["request_context"]

    # Request body
    controller_details = await request.json()

    # Initialise MyData DID Manager.
    mgr = V2ADAManager(context=context)

    # Call the function
    record = await mgr.update_controller_details(
        organisation_name=controller_details.get("organisation_name"),
        cover_image_url=controller_details.get("cover_image_url"),
        logo_image_url=controller_details.get("logo_image_url"),
        location=controller_details.get("location"),
        organisation_type=controller_details.get("organisation_type"),
        description=controller_details.get("description"),
        policy_url=controller_details.get("policy_url"),
        eula_url=controller_details.get("eula_url")
    )

    return web.json_response(record.serialize())



import json
import uuid
import logging
from aiohttp import web
from aiohttp_apispec import (
    docs,
    request_schema,
    querystring_schema,
    response_schema,
    match_info_schema,
)
from aries_cloudagent.messaging.models.base import BaseModelError
from aries_cloudagent.connections.models.connection_record import (
    ConnectionRecord,
)
from aries_cloudagent.protocols.connections.v1_0.manager import (
    ConnectionManager,
    ConnectionManagerError,
)
from dexa_sdk.managers.ada_manager import V2ADAManager
from dexa_sdk.utils import clean_and_get_field_from_dict
from ..utils.util import str_to_bool
from ..manager import ADAManagerError
from .openapi import (
    V2CreateInvitationQueryStringSchema,
    V2InvitationResultSchema,
    GenerateFirebaseDynamicLinkForConnectionInvitationMatchInfoSchema,
    GenerateFirebaseDynamicLinkForConnectionInvitationResponseSchema,
    SendExistingConnectionsMessageHandlerMatchInfoSchema,
    SendExistingConnectionsMessageHandlerRequestSchema,
    GetExistingConnectionMatchInfoSchema,
    GetExistingConnectionResponseSchema,
    ConnectionsListQueryStringSchemaV2,
    ConnectionListSchema,
)

LOGGER = logging.getLogger(__name__)

PAGINATION_PAGE_SIZE = 10


@docs(
    tags=["connection"],
    summary="Well-known endpoint for connection",
)
async def wellknown_connection_handler(request: web.BaseRequest):
    """Handler for well-known connection for the agent."""
    context = request.app["request_context"]
    auto_accept = True
    alias = ""
    public = False
    multi_use = False

    connection_mgr = ConnectionManager(context)
    try:
        (connection, invitation) = await connection_mgr.create_invitation(
            auto_accept=auto_accept, public=public, multi_use=multi_use, alias=alias
        )

        result = {
            "ServiceEndpoint": invitation.serialize()["serviceEndpoint"],
            "RoutingKey": "",
            "Invitation": {
                "label": invitation.label,
                "serviceEndpoint": invitation.serialize()["serviceEndpoint"],
                "routingKeys": [],
                "recipientKeys": invitation.serialize()["recipientKeys"],
                "@id": str(uuid.uuid4()),
                "@type": "did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/connections/1.0/invitation",
            },
        }
    except (ConnectionManagerError, BaseModelError) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response(result)


@docs(
    tags=["connection"],
    summary="Create a new connection invitation (Overridden API)",
)
@querystring_schema(V2CreateInvitationQueryStringSchema())
@response_schema(V2InvitationResultSchema(), 200)
async def v2_connections_create_invitation(request: web.BaseRequest):
    """
    Request handler for creating a new connection invitation.

    Args:
        request: aiohttp request object

    Returns:
        The connection invitation details

    """

    context = request.app["request_context"]
    auto_accept = json.loads(request.query.get("auto_accept", "null"))
    alias = request.query.get("alias")
    public = json.loads(request.query.get("public", "false"))
    multi_use = json.loads(request.query.get("multi_use", "false"))

    if public and not context.settings.get("public_invites"):
        raise web.HTTPForbidden(
            reason="Configuration does not include public invitations"
        )
    base_url = context.settings.get("invite_base_url")

    # Initialise MyData DID Manager.
    mgr = V2ADAManager(context=context)
    try:
        (connection, invitation) = await mgr.create_invitation(
            auto_accept=auto_accept, public=public, multi_use=multi_use, alias=alias
        )

        result = {
            "connection_id": connection and connection.connection_id,
            "invitation": invitation.serialize(),
            "invitation_url": invitation.to_url(base_url),
        }
    except (ConnectionManagerError, BaseModelError) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    if connection and connection.alias:
        result["alias"] = connection.alias

    return web.json_response(result)


@docs(
    tags=["connection"],
    summary="Generate firebase dynamic link for connection invitation",
)
@match_info_schema(GenerateFirebaseDynamicLinkForConnectionInvitationMatchInfoSchema())
@response_schema(
    GenerateFirebaseDynamicLinkForConnectionInvitationResponseSchema(), 200
)
async def generate_firebase_dynamic_link_for_connection_invitation_handler(
    request: web.BaseRequest,
):
    """
    Request handler for generating firebase dynamic link for connection invitation.

    Args:
        request: aiohttp request object

    Returns:
        The firebase dynamic link

    """

    # Request context.
    context = request.app["request_context"]

    # Path params.
    conn_id = request.match_info["conn_id"]

    # Initialise MyData DID Manager.
    mgr = V2ADAManager(context=context)
    try:
        # Call the function
        res = await mgr.create_connection_qr_code(
            conn_id
        )

    except (ConnectionManagerError, BaseModelError) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response(res)


@docs(
    tags=["connection"],
    summary="Send existing connections message to remote agent.",
    responses={
        200: {
            "description": "Success",
        }
    },
)
@match_info_schema(SendExistingConnectionsMessageHandlerMatchInfoSchema())
@request_schema(SendExistingConnectionsMessageHandlerRequestSchema())
async def send_existing_connections_message_handler(request: web.BaseRequest):
    """Send existing connections message to remote agent."""

    context = request.app["request_context"]
    conn_id = request.match_info["conn_id"]

    # Fetch request body
    body = await request.json()
    theirdid = body.get("theirdid")

    # Initialise MyData DID Manager.
    mgr = V2ADAManager(context=context)

    try:
        # Call the function
        await mgr.send_existing_connections_message(
            theirdid,
            conn_id
        )

    except ADAManagerError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response({}, status=200)


@docs(
    tags=["connection"],
    summary="Fetch existing connection details if any for a current connection.",
)
@match_info_schema(GetExistingConnectionMatchInfoSchema())
@response_schema(GetExistingConnectionResponseSchema(), 200)
async def get_existing_connections_handler(request: web.BaseRequest):
    """Fetch existing connection details if any for a current connection."""

    # Request context
    context = request.app["request_context"]

    # Path params.
    conn_id = request.match_info["conn_id"]

    # Initialise MyData DID Manager.
    mgr = V2ADAManager(context=context)

    # Call the function
    result = await mgr.get_existing_connection_record_for_new_connection_id(
        conn_id
    )
    result = result.serialize() if result else {}

    return web.json_response(result)


def connection_sort_key(conn):
    """Get the sorting key for a particular connection."""
    if conn["state"] == ConnectionRecord.STATE_INACTIVE:
        pfx = "2"
    elif conn["state"] == ConnectionRecord.STATE_INVITATION:
        pfx = "1"
    else:
        pfx = "0"
    return pfx + conn["created_at"]


@docs(
    tags=["connection"],
    summary="Query agent-to-agent connections (v2)",
)
@querystring_schema(ConnectionsListQueryStringSchemaV2())
@response_schema(ConnectionListSchema(), 200)
async def connections_list_v2(request: web.BaseRequest):
    """
    Request handler for searching connection records.

    Args:
        request: aiohttp request object

    Returns:
        The connection list response

    """
    context = request.app["request_context"]
    tag_filter = {}
    for param_name in (
        "invitation_id",
        "my_did",
        "their_did",
        "request_id",
        "invitation_key"
    ):
        if param_name in request.query and request.query[param_name] != "":
            tag_filter[param_name] = request.query[param_name]
    post_filter = {}
    for param_name in (
        "alias",
        "initiator",
        "state",
        "their_role",
    ):
        if param_name in request.query and request.query[param_name] != "":
            post_filter[param_name] = request.query[param_name]

    # Pagination parameters
    page = clean_and_get_field_from_dict(request.query, "page")
    page = int(page) if page is not None else page
    page_size = clean_and_get_field_from_dict(request.query, "page_size")
    page_size = int(page_size) if page_size is not None else page_size

    # Category
    org_flag = clean_and_get_field_from_dict(request.query, "org_flag")
    marketplace_flag = clean_and_get_field_from_dict(request.query, "marketplace_flag")

    mgr = V2ADAManager(context)

    pagination_result = await mgr.query_connections_and_categorise_results(
        tag_filter,
        post_filter,
        page if page else 1,
        page_size if page_size else 10,
        str_to_bool(org_flag) if org_flag else org_flag,
        str_to_bool(marketplace_flag) if marketplace_flag else marketplace_flag
    )

    return web.json_response(pagination_result._asdict())




import json
import typing
from typing import Any

from marshmallow import fields, validate

from aries_cloudagent.messaging.models.base_record import BaseExchangeRecord, BaseExchangeSchema, match_post_filter
from aries_cloudagent.messaging.valid import UUIDFour
from aries_cloudagent.config.injection_context import InjectionContext
from aries_cloudagent.storage.base import BaseStorage, StorageDuplicateError, StorageNotFoundError

from ...utils.util import bool_to_str

class DataAgreementV1Record(BaseExchangeRecord):
    """
    DataAgreementV1Record model class for serialisation/deserialisation of data agreement records to and from wallet.
    Data agreement schema will be version 1.0.
    """
    class Meta:
        # Data Agreement Record schema (version 1.0)
        schema_class = "DataAgreementV1RecordSchema"

    # Wallet record type
    RECORD_TYPE = "data_agreement_record"

    # Wallet record identifier field
    RECORD_ID_NAME = "data_agreement_record_id"

    # Webhook topic name for this record type
    WEBHOOK_TOPIC = None

    # Wallet record tags used for filtering
    # Note: These are not tags for the ledger, but rather tags for the wallet
    #      to group records.
    TAG_NAMES = {
        "~method_of_use", 
        "~data_agreement_id", 
        "~publish_flag", 
        "~delete_flag",
        "~schema_id",
        "~cred_def_id",
        "~is_existing_schema"
    }

    # State of the data agreement.
    # Only one possible value at this stage of the DA - preparation
    STATE_PREPARATION = "PREPARATION"

    METHOD_OF_USE_DATA_SOURCE = "data-source"
    METHOD_OF_USE_DATA_USING_SERVICE = "data-using-service"

    def __init__(
        self,
        *,
        data_agreement_record_id: str = None,
        data_agreement_id: str = None,
        state: str = None,
        method_of_use: str = None,
        data_agreement: dict = None,
        publish_flag: str = "false",
        delete_flag: str = "false",
        schema_id: str = None,
        cred_def_id: str = None,
        data_agreement_proof_presentation_request: dict = None,
        is_existing_schema: str = "false",
        **kwargs
    ):
        """
        Initialise a new DataAgreementRecordV1 instance.

        Args:
            data_agreement_id: The unique identifier for the data agreement.
            state: The state of the data agreement.
            method_of_use: The method of use for the data agreement.
            data_agreement: The data agreement.
            kwargs: Any other parameters.
        """
        super().__init__(data_agreement_record_id, state, **kwargs)
        self.method_of_use = method_of_use
        self.state = state
        self.data_agreement = data_agreement
        self.data_agreement_id = data_agreement_id
        self.publish_flag = publish_flag
        self.delete_flag = delete_flag
        self.schema_id = schema_id
        self.cred_def_id = cred_def_id
        self.data_agreement_proof_presentation_request = data_agreement_proof_presentation_request
        self.is_existing_schema = is_existing_schema

    @property
    def data_agreement_record_id(self) -> str:
        """Accessor for data_agreement_record_id."""
        return self._id

    @property
    def record_value(self) -> dict:
        """Accessor for JSON record value generated for this transaction record."""
        return {
            prop: getattr(self, prop)
            for prop in (
                "state",
                "method_of_use",
                "data_agreement",
                "data_agreement_id",
                "publish_flag",
                "delete_flag",
                "schema_id",
                "cred_def_id",
                "data_agreement_proof_presentation_request",
                "is_existing_schema",
            )
        }

    def __eq__(self, other: Any) -> bool:
        """Comparison between records."""
        return super().__eq__(other)
    
    @property
    def _publish_flag(self) -> bool:
        """Accessor for publish_flag."""
        return self.publish_flag == "true"

    @_publish_flag.setter
    def _publish_flag(self, value: bool) -> None:
        """Setter for publish_flag."""
        self.publish_flag = "true" if value else "false"
    
    @property
    def _delete_flag(self) -> bool:
        """Accessor for delete_flag."""
        return self.delete_flag == "true"
    
    @_delete_flag.setter
    def _delete_flag(self, value: bool) -> None:
        """Setter for delete_flag."""
        self.delete_flag = "true" if value else "false"
    
    @property
    def _is_existing_schema(self) -> bool:
        """Accessor for is_existing_schema."""
        return self.is_existing_schema == "true"
    
    @_is_existing_schema.setter
    def _is_existing_schema(self, value: bool) -> None:
        """Setter for is_existing_schema."""
        self.is_existing_schema = "true" if value else "false"

    @property
    def is_published(self) -> bool:
        """Check if data agreement record is published."""
        return True if self._publish_flag and not self._delete_flag else False
    
    @property
    def is_deleted(self) -> bool:
        """Check if data agreemnent is deleted."""
        return True if self._delete_flag and not self._publish_flag else False

    @property
    def is_draft(self) -> bool:
        """Check if data agreement is a draft."""
        return True if not self._publish_flag and not self._delete_flag else False
    
    @classmethod
    async def retrieve_non_deleted_data_agreement_by_id(
        cls, 
        context: InjectionContext,
        data_agreement_id: str,
    ) -> "DataAgreementV1Record":
        """
        Retrieve a non-deleted data agreement record by its data agreement id.

        Args:
            context: The injection context to use.
            data_agreement_id: The data agreement id.
        
        Returns:
            The data agreement record.
        """

        tag_filter: dict = {
            "data_agreement_id": data_agreement_id,
            "delete_flag": bool_to_str(False)
        }
        post_filter: dict = None

        return await cls.retrieve_by_tag_filter(
            context,
            tag_filter,
            post_filter
        )
    
    @classmethod
    async def retrieve_all_non_deleted_data_agreements(
        cls,
        context: InjectionContext,
    ) -> typing.List["DataAgreementV1Record"]:
        """
        Retrieve all non-deleted data agreements.

        Args:
            context: The injection context to use.

        Returns:
            The data agreements.
        """

        tag_filter: dict = {
            "delete_flag": bool_to_str(False)
        }

        return await cls.query(
            context,
            tag_filter=tag_filter,
        )

class DataAgreementV1RecordSchema(BaseExchangeSchema):

    class Meta:
        """DataAgreementRecordV1Schema metadata."""

        # Model class
        model_class = DataAgreementV1Record

    # Data agreement record identifier
    data_agreement_record_id = fields.Str(
        required=True,
        description="Data Agreement Record identifier",
        example=UUIDFour.EXAMPLE
    )

    # Data agreement identifier
    data_agreement_id = fields.Str(
        required=True,
        description="The unique identifier for the data agreement.",
        example=UUIDFour.EXAMPLE
    )

    # State of the data agreement.
    state = fields.Str(
        required=True,
        description="The state of the data agreement.",
        example=DataAgreementV1Record.STATE_PREPARATION,
        validate=validate.OneOf(
            [
                DataAgreementV1Record.STATE_PREPARATION,
            ]
        )
    )

    # Method of use for the data agreement.
    method_of_use = fields.Str(
        required=True,
        description="The method of use for the data agreement.",
        example="data-source",
        validate=validate.OneOf(
            [
                DataAgreementV1Record.METHOD_OF_USE_DATA_SOURCE,
                DataAgreementV1Record.METHOD_OF_USE_DATA_USING_SERVICE,
            ]
        )
    )

    # Data agreement
    data_agreement = fields.Dict(
        required=True,
        description="The data agreement.",
    )

    # Production flag
    publish_flag = fields.Str(
        required=True,
        description="The production flag.",
        example="false",
        validate=validate.OneOf(
            [
                "true",
                "false",
            ]
        )
    )

    # Delete flag
    delete_flag = fields.Str(
        required=True,
        description="The delete flag.",
        example="false",
        validate=validate.OneOf(
            [
                "true",
                "false",
            ]
        )
    )

    # Schema identifier
    schema_id = fields.Str(
        required=True,
        description="The schema identifier.",
        example="WgWxqztrNooG92RXvxSTWv:2:schema_name:1.0"
    )

    # Credential definition identifier
    cred_def_id = fields.Str(
        required=True,
        description="The credential definition identifier.",
        example="WgWxqztrNooG92RXvxSTWv:3:CL:20:tag"
    )

    # Data agreement proof presentation request
    data_agreement_proof_presentation_request = fields.Dict(
        required=True,
        description="The data agreement proof presentation request.",
    )

    is_existing_schema = fields.Str(
        required=True,
        description="Is existing schema.",
        example="false",
        validate=validate.OneOf(
            [
                "true",
                "false",
            ]
        )
    )



from enum import Enum
from marshmallow import EXCLUDE, fields, validate

from aries_cloudagent.messaging.agent_message import AgentMessage, AgentMessageSchema
from aries_cloudagent.messaging.valid import UUIDFour

from ..message_types import (
    DATA_AGREEMENT_PROBLEM_REPORT,
    MYDATA_DID_PROBLEM_REPORT,
    PROTOCOL_PACKAGE,
    DATA_AGREEMENT_NEGOTIATION_PROBLEM_REPORT,
    DATA_AGREEMENT_TERMINATION_PROBLEM_REPORT
)
from ..utils.regex import MYDATA_DID

# Handler class
HANDLER_CLASS = (
    f"{PROTOCOL_PACKAGE}.handlers"
    ".mydata_did_problem_report_handler.MyDataDIDProblemReportHandler"
)


class MyDataDIDProblemReportMessageReason(str, Enum):
    """Supported reason codes for mydata-did message family."""

    # MyData DID exists.
    DID_EXISTS = "mydata_did_exists"

    # DIDComm message body signature verification failed.
    MESSAGE_BODY_SIGNATURE_VERIFICATION_FAILED = "signature_verification_failed"

    # MyData DID not found.
    DID_NOT_FOUND = "mydata_did_not_found"

    # MyData DID revoked.
    DID_REVOKED = "mydata_did_revoked"


class MyDataDIDProblemReportMessage(AgentMessage):
    """
    Problem report message for MyData DID protocol.
    """

    class Meta:
        # Handler class that should handle this message
        handler_class = HANDLER_CLASS

        # Message type for this message
        message_type = MYDATA_DID_PROBLEM_REPORT

        # Schema for this message
        schema_class = "MyDataDIDProblemReportMessageSchema"

    def __init__(
        self,
        *,
        problem_code: str = None,
        explain: str = None,
        from_did: str = None,
        to_did: str = None,
        created_time: str = None,
        **kwargs
    ):
        """
        Initialize mydata-did problem report message instance.

        Args:
            problem_code: Reason code for the problem.
            explain: Explanation for the problem.
            from_did: DID of the sender.
            to_did: DID of the receiver.
            created_time: Time when the problem report was created.
        """

        super().__init__(**kwargs)

        # Set attributes
        self.explain = explain
        self.problem_code = problem_code
        self.from_did = from_did
        self.to_did = to_did
        self.created_time = created_time


class MyDataDIDProblemReportMessageSchema(AgentMessageSchema):
    """
    Schema for mydata-did problem report message.
    """

    class Meta:
        # Message class this schema is derived from.
        model_class = MyDataDIDProblemReportMessage

        # Unknow fields are excluded.
        unknown = EXCLUDE

    # From DID
    from_did = fields.Str(data_key="from", **MYDATA_DID)

    # To DID
    to_did = fields.Str(data_key="to", **MYDATA_DID)

    # Created time
    created_time = fields.Str(data_key="created_time")

    # Explaination for the problem
    explain = fields.Str(
        required=False,
        description="Localized error explanation",
        example="Invitation not accepted",
    )

    # Reason code for the problem
    problem_code = fields.Str(
        data_key="problem-code",
        required=False,
        description="Standard error identifier",
        validate=validate.OneOf(
            choices=[prr.value for prr in MyDataDIDProblemReportMessageReason],
            error="Value {input} must be one of {choices}.",
        ),
        example=MyDataDIDProblemReportMessageReason.DID_EXISTS.value,
    )


# Handler class path for Data Agreement Problem Report (data-agreements/1.0/problem-report) message
DATA_AGREEMENT_PROBLEM_REPORT_HANDLER_CLASS = (
    f"{PROTOCOL_PACKAGE}.handlers"
    ".data_agreement_problem_report_handler.DataAgreementProblemReportHandler"
)


class DataAgreementProblemReportReason(str, Enum):
    """Supported reason codes."""

    # Triggered when data agreement was not found.
    DATA_AGREEMENT_NOT_FOUND = "data_agreement_not_found"

    # Triggered when read-data-agreement failed due to processing error.
    READ_DATA_AGREEMENT_FAILED = "read_data_agreement_failed"


class DataAgreementProblemReport(AgentMessage):
    """Base class representing a data agreement problem report message."""

    class Meta:
        """Data agreement problem report metadata."""

        handler_class = DATA_AGREEMENT_PROBLEM_REPORT_HANDLER_CLASS
        message_type = DATA_AGREEMENT_PROBLEM_REPORT
        schema_class = "DataAgreementProblemReportSchema"

    def __init__(
        self,
        *,
        problem_code: str = None,
        explain: str = None,
        from_did: str = None,
        to_did: str = None,
        created_time: str = None,
        **kwargs
    ):
        """
        Initialize a DataAgreementProblemReport message instance.
        """
        super().__init__(**kwargs)
        self.explain = explain
        self.problem_code = problem_code
        self.from_did = from_did
        self.to_did = to_did
        self.created_time = created_time


class DataAgreementProblemReportSchema(AgentMessageSchema):

    class Meta:
        """Metadata for data agreement problem report schema."""

        model_class = DataAgreementProblemReport
        unknown = EXCLUDE

    from_did = fields.Str(data_key="from", **MYDATA_DID)
    to_did = fields.Str(data_key="to", **MYDATA_DID)
    created_time = fields.Str(data_key="created_time")
    explain = fields.Str(
        required=False,
        description="Localized error explanation",
        example="Invitation not accepted",
    )
    problem_code = fields.Str(
        data_key="problem-code",
        required=False,
        description="Standard error identifier",
        validate=validate.OneOf(
            choices=[dapr.value for dapr in DataAgreementProblemReportReason],
            error="Value {input} must be one of {choices}.",
        ),
        example=DataAgreementProblemReportReason.DATA_AGREEMENT_NOT_FOUND.value,
    )


# Handler class path for Data Agreement Negotiation Problem Report (data-agreement-negotiation/1.0/problem-report) message
DATA_AGREEMENT_NEGOTIATION_PROBLEM_REPORT_HANDLER_CLASS = (
    f"{PROTOCOL_PACKAGE}.handlers"
    ".data_agreement_negotiation_problem_report_handler.DataAgreementNegotiationProblemReportHandler"
)


class DataAgreementNegotiationProblemReportReason(str, Enum):
    """Supported reason codes."""

    # Trigger when data agreement signature verification fails
    SIGNATURE_VERIFICATION_FAILED = "signature_verification_failed"

    # Trigger when controller (Organisation) DID is not present in remote did registry
    CONTROLLER_DID_INVALID = "controller_did_invalid"

    # Trigger when principle (Data Subject) DID is invalid
    PRINCIPLE_DID_INVALID = "principle_did_invalid"

    # Trigger when data agreement context is invalid
    DATA_AGREEMENT_CONTEXT_INVALID = "data_agreement_context_invalid"


class DataAgreementNegotiationProblemReport(AgentMessage):
    """Base class representing a data agreement negotiation problem report message."""

    class Meta:
        """Data agreement negotiation problem report metadata."""

        handler_class = DATA_AGREEMENT_NEGOTIATION_PROBLEM_REPORT_HANDLER_CLASS
        message_type = DATA_AGREEMENT_NEGOTIATION_PROBLEM_REPORT
        schema_class = "DataAgreementNegotiationProblemReportSchema"

    def __init__(
        self,
        *,
        problem_code: str = None,
        explain: str = None,
        from_did: str = None,
        to_did: str = None,
        created_time: str = None,
        data_agreement_id: str = None,
        **kwargs
    ):
        """
        Initialize a DataAgreementNegotiationProblemReport message instance.

        Args:
            explain: The localized error explanation
            problem_code: The standard error identifier
            from_did: Sender DID
            to_did: Receipient DID
            created_time: The timestamp of the message
            data_agreement_id: The data agreement identifier
        """
        super().__init__(**kwargs)
        self.explain = explain
        self.problem_code = problem_code
        self.from_did = from_did
        self.to_did = to_did
        self.created_time = created_time
        self.data_agreement_id = data_agreement_id


class DataAgreementNegotiationProblemReportSchema(AgentMessageSchema):
    """
    Data agreement negotiation problem report schema.
    """
    class Meta:
        """Metadata for data agreement negotiation problem report schema."""

        model_class = DataAgreementNegotiationProblemReport
        unknown = EXCLUDE

    from_did = fields.Str(data_key="from", **MYDATA_DID)
    to_did = fields.Str(data_key="to", **MYDATA_DID)
    created_time = fields.Str(data_key="created_time")
    explain = fields.Str(
        required=False,
        description="Localized error explanation",
        example="Invitation not accepted",
    )
    problem_code = fields.Str(
        data_key="problem-code",
        required=False,
        description="Standard error identifier",
        validate=validate.OneOf(
            choices=[
                dapr.value for dapr in DataAgreementNegotiationProblemReportReason],
            error="Value {input} must be one of {choices}.",
        ),
        example=DataAgreementNegotiationProblemReportReason.SIGNATURE_VERIFICATION_FAILED.value,
    )
    data_agreement_id = fields.Str(
        data_key="data-agreement-id",
        required=False,
        description="The data agreement identifier",
        example=UUIDFour.EXAMPLE,
    )


# Handler class path for Data Agreement Termination Problem Report (data-agreement-termination/1.0/problem-report) message
DATA_AGREEMENT_TERMINATION_PROBLEM_REPORT_HANDLER_CLASS = (
    f"{PROTOCOL_PACKAGE}.handlers"
    ".data_agreement_termination_problem_report_handler.DataAgreementTerminationProblemReportHandler"
)


class DataAgreementTerminationProblemReportReason(str, Enum):
    """Supported reason codes."""

    # Trigger when data agreement signature verification fails.
    SIGNATURE_VERIFICATION_FAILED = "signature_verification_failed"

    # Trigger when data agreement is not found.
    DATA_AGREEMENT_NOT_FOUND = "data_agreement_not_found"

    # Trigger when principle (Data Subject) DID is invalid.
    PRINCIPLE_DID_INVALID = "principle_did_invalid"


class DataAgreementTerminationProblemReport(AgentMessage):
    """Base class representing a data agreement termination problem report message."""

    class Meta:
        """Data agreement termination problem report metadata."""

        handler_class = DATA_AGREEMENT_TERMINATION_PROBLEM_REPORT_HANDLER_CLASS
        message_type = DATA_AGREEMENT_TERMINATION_PROBLEM_REPORT
        schema_class = "DataAgreementTerminationProblemReportSchema"

    def __init__(
        self,
        *,
        problem_code: str = None,
        explain: str = None,
        from_did: str = None,
        to_did: str = None,
        created_time: str = None,
        data_agreement_id: str = None,
        **kwargs
    ):
        """
        Initialize a DataAgreementTerminationProblemReport message instance.

        Args:
            explain: The localized error explanation
            problem_code: The standard error identifier
            from_did: Sender DID
            to_did: Receipient DID
            created_time: The timestamp of the message
            data_agreement_id: The data agreement identifier
        """
        super().__init__(**kwargs)
        self.explain = explain
        self.problem_code = problem_code
        self.from_did = from_did
        self.to_did = to_did
        self.created_time = created_time
        self.data_agreement_id = data_agreement_id


class DataAgreementTerminationProblemReportSchema(AgentMessageSchema):
    """
    Data agreement termination problem report schema.
    """
    class Meta:
        """Metadata for data agreement termination problem report schema."""

        model_class = DataAgreementTerminationProblemReport
        unknown = EXCLUDE

    from_did = fields.Str(data_key="from", **MYDATA_DID)
    to_did = fields.Str(data_key="to", **MYDATA_DID)
    created_time = fields.Str(data_key="created_time")
    explain = fields.Str(
        required=False,
        description="Localized error explanation",
        example="Invitation not accepted",
    )
    problem_code = fields.Str(
        data_key="problem-code",
        required=False,
        description="Standard error identifier",
        validate=validate.OneOf(
            choices=[
                dapr.value for dapr in DataAgreementTerminationProblemReportReason],
            error="Value {input} must be one of {choices}.",
        ),
        example=DataAgreementTerminationProblemReportReason.SIGNATURE_VERIFICATION_FAILED.value,
    )
    data_agreement_id = fields.Str(
        data_key="data-agreement-id",
        required=False,
        description="The data agreement identifier",
        example=UUIDFour.EXAMPLE,
    )


from aries_cloudagent.messaging.base_handler import BaseHandler, BaseResponder, RequestContext, HandlerException
from aries_cloudagent.storage.record import StorageRecord
from aries_cloudagent.wallet.base import BaseWallet
from aries_cloudagent.wallet.indy import IndyWallet

from ..messages.data_agreement_terminate import DataAgreementTerminationTerminateMessage
from ..messages.data_agreement_terminate_ack import DataAgreementTerminationAck
from ..manager import ADAManager
from ..models.data_agreement_termination_terminate_model import DataAgreementTerminationTerminateBody
from ..models.exchange_records.data_agreement_record import DataAgreementV1Record
from ..models.data_agreement_instance_model import DataAgreementInstance, DataAgreementInstanceSchema
from ..utils.did.mydata_did import DIDMyData
from ..utils.jsonld.data_agreement import verify_data_agreement
from ..messages.problem_report import (
    DataAgreementTerminationProblemReport,
    DataAgreementTerminationProblemReportReason
)

from ...patched_protocols.issue_credential.v1_0.models.credential_exchange import (
    V10CredentialExchange
)

from ...patched_protocols.present_proof.v1_0.models.presentation_exchange import (
    V10PresentationExchange
)

import json
import datetime


class DataAgreementTerminationTerminateMessageHandler(BaseHandler):
    """Handler for data-agreement-termination/1.0/terminate message."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Message handler logic for data-agreement-termination/1.0/terminate message."""

        # Assert that the message is of the correct type
        assert isinstance(
            context.message, DataAgreementTerminationTerminateMessage)

        self._logger.info(
            "Received data-agreement-termination/1.0/terminate message: \n%s",
            json.dumps(context.message.serialize(), indent=4)
        )

        # Check if connection is ready
        if not context.connection_ready:
            self._logger.info(
                "Connection not active, skipping data-agreement-termination/1.0/terminate handler: %s",
                context.message_receipt.sender_did,
            )
            return

        data_agreement_termination_terminate_message = context.message
        data_agreement_termination_terminate_message_body: DataAgreementTerminationTerminateBody = data_agreement_termination_terminate_message.body

        # Wallet instance from request context
        wallet: IndyWallet = await context.inject(BaseWallet)

        # Initialize ADA manager
        ada_manager = ADAManager(context)

        # Fetch the data agreement instance metadata
        data_agreement_instance_metadata_records = await ada_manager.query_data_agreement_instance_metadata(
            tag_query={
                'data_agreement_id': data_agreement_termination_terminate_message_body.data_agreement_id,
            }
        )

        # Check if there is a data agreement instance metadata record
        if not data_agreement_instance_metadata_records:
            self._logger.info(
                "Data agreement not found; Failed to handle terminate message for data agreement: %s",
                data_agreement_termination_terminate_message_body.data_agreement_id,
            )
            return

        if len(data_agreement_instance_metadata_records) > 1:
            self._logger.info(
                "Duplicate data agreement records found; Failed to handle terminate message for data agreement: %s",
                data_agreement_termination_terminate_message_body.data_agreement_id,
            )
            return

        data_agreement_instance_metadata_record: StorageRecord = data_agreement_instance_metadata_records[
            0]

        # Identify the method of use

        if data_agreement_instance_metadata_record.tags.get("method_of_use") == DataAgreementV1Record.METHOD_OF_USE_DATA_SOURCE:

            # Fetch exchante record (credential exchange if method of use is "data-source")
            tag_filter = {}
            post_filter = {
                "data_agreement_id": data_agreement_termination_terminate_message_body.data_agreement_id
            }
            records = await V10CredentialExchange.query(context, tag_filter, post_filter)

            if not records:
                self._logger.info(
                    "Credential exchange record not found; Failed to handle terminate message for data agreement: %s",
                    data_agreement_termination_terminate_message_body.data_agreement_id,
                )
                return

            if len(records) > 1:
                self._logger.info(
                    "Duplicate credential exchange records found; Failed to handle terminate message for data agreement: %s",
                    data_agreement_termination_terminate_message_body.data_agreement_id,
                )
                return

            cred_ex_record: V10CredentialExchange = records[0]

            # Check if data agreement is in "accept" status
            if cred_ex_record.data_agreement_status != V10CredentialExchange.DATA_AGREEMENT_ACCEPT:
                self._logger.info(
                    "Credential exchange record not in offer sent state; Failed to handle terminate message for data agreement: %s",
                    data_agreement_termination_terminate_message_body.data_agreement_id,
                )
                return

            # Reconstruct the data agreement

            # Deserialise data agreement
            data_agreement_instance: DataAgreementInstance = DataAgreementInstanceSchema().load(
                cred_ex_record.data_agreement
            )

            # Check if terminate message is signed by data agreement principle did
            if data_agreement_instance.principle_did != data_agreement_termination_terminate_message_body.proof.verification_method:
                self._logger.info(
                    "Data agreement principle DID does not match sender DID; Failed to handle terminate message for data agreement: %s",
                    data_agreement_termination_terminate_message_body.data_agreement_id,
                )

                # Send problem report.

                problem_report = DataAgreementTerminationProblemReport(
                    from_did=data_agreement_termination_terminate_message.to_did,
                    to_did=data_agreement_termination_terminate_message.from_did,
                    created_time=str(
                        int(datetime.datetime.utcnow().timestamp())),
                    problem_code=DataAgreementTerminationProblemReportReason.PRINCIPLE_DID_INVALID.value,
                    explain=f"Data agreement principle DID does not match sender DID; Failed to process terminate message for data agreement: {data_agreement_termination_terminate_message.body.data_agreement_id}",
                    data_agreement_id=data_agreement_termination_terminate_message_body.data_agreement_id
                )

                problem_report.assign_thread_id(
                    thid=data_agreement_termination_terminate_message._id
                )

                # Update credential exchange record with data agreement metadata
                cred_ex_record.data_agreement_problem_report = problem_report.serialize()
                cred_ex_record.data_agreement_status = V10PresentationExchange.DATA_AGREEMENT_PROBLEM_REPORT
                await cred_ex_record.save(context)

                await responder.send_reply(problem_report)

                return

            # Update data agreement event with terminate event
            data_agreement_instance.event.append(
                data_agreement_termination_terminate_message_body.event
            )

            # Update data agreement proof chain with terminate proof
            data_agreement_instance.proof_chain.append(
                data_agreement_termination_terminate_message_body.proof
            )

            # Verify signatures on data agreement
            verkeys = []

            for event in data_agreement_instance.event:
                temp_verkey = DIDMyData.from_did(event.did).public_key_b58
                verkeys.append(temp_verkey)

            valid = await verify_data_agreement(
                data_agreement_instance.serialize(),
                verkeys[-1],
                wallet,
                drop_proof_chain=False
            )

            if not valid:
                self._logger.error(
                    "Data agreement terminate verification failed"
                )

                # Send problem report

                problem_report = DataAgreementTerminationProblemReport(
                    from_did=data_agreement_termination_terminate_message.to_did,
                    to_did=data_agreement_termination_terminate_message.from_did,
                    created_time=str(
                        int(datetime.datetime.utcnow().timestamp())),
                    problem_code=DataAgreementTerminationProblemReportReason.SIGNATURE_VERIFICATION_FAILED.value,
                    explain=f"Data agreement terminate verification failed; Failed to process terminate message for data agreement: {data_agreement_termination_terminate_message.body.data_agreement_id}",
                    data_agreement_id=data_agreement_termination_terminate_message_body.data_agreement_id
                )

                problem_report.assign_thread_id(
                    thid=data_agreement_termination_terminate_message._id
                )

                # Update credential exchange record with data agreement metadata
                cred_ex_record.data_agreement_problem_report = problem_report.serialize()
                cred_ex_record.data_agreement_status = V10PresentationExchange.DATA_AGREEMENT_PROBLEM_REPORT
                await cred_ex_record.save(context)

                await responder.send_reply(problem_report)

                raise HandlerException(
                    "Data agreement terminate signature verification failed"
                )

            # Update credential exchange record with data agreement metadata
            cred_ex_record.data_agreement = data_agreement_instance.serialize()
            cred_ex_record.data_agreement_status = V10CredentialExchange.DATA_AGREEMENT_TERMINATE

            await cred_ex_record.save(context)

            # Construct terminate ack message
            data_agreement_terminate_ack = DataAgreementTerminationAck(
                status="TERMINATE OK"
            )

            data_agreement_terminate_ack.assign_thread_id(
                thid=data_agreement_termination_terminate_message._id
            )

            await responder.send_reply(data_agreement_terminate_ack)

        if data_agreement_instance_metadata_record.tags.get("method_of_use") == DataAgreementV1Record.METHOD_OF_USE_DATA_USING_SERVICE:

            # Fetch exchange record (presentation exchange if method of use is "data-using-service")
            tag_filter = {}
            post_filter = {
                "data_agreement_id": data_agreement_termination_terminate_message_body.data_agreement_id
            }
            records = await V10PresentationExchange.query(context, tag_filter, post_filter)

            if not records:
                self._logger.info(
                    "Presentation exchange record not found; Failed to handle terminate message for data agreement: %s",
                    data_agreement_termination_terminate_message_body.data_agreement_id,
                )
                return

            if len(records) > 1:
                self._logger.info(
                    "Duplicate presentation exchange records found; Failed to handle terminate message for data agreement: %s",
                    data_agreement_termination_terminate_message_body.data_agreement_id,
                )
                return

            pres_ex_record: V10PresentationExchange = records[0]

            # Check if data agreement is in "accept" status
            if pres_ex_record.data_agreement_status != V10PresentationExchange.DATA_AGREEMENT_ACCEPT:
                self._logger.info(
                    "Presentation exchange record not in offer sent state; Failed to handle terminate message for data agreement: %s",
                    data_agreement_termination_terminate_message_body.data_agreement_id,
                )
                return

            # Reconstruct the data agreement

            # Deserialise data agreement
            data_agreement_instance: DataAgreementInstance = DataAgreementInstanceSchema().load(
                pres_ex_record.data_agreement
            )

            # Check if terminate message is signed by data agreement principle did
            if data_agreement_instance.principle_did != data_agreement_termination_terminate_message_body.proof.verification_method:
                self._logger.info(
                    "Data agreement principle DID does not match sender DID; Failed to handle terminate message for data agreement: %s",
                    data_agreement_termination_terminate_message_body.data_agreement_id,
                )

                # Send problem report.

                problem_report = DataAgreementTerminationProblemReport(
                    from_did=data_agreement_termination_terminate_message.to_did,
                    to_did=data_agreement_termination_terminate_message.from_did,
                    created_time=str(
                        int(datetime.datetime.utcnow().timestamp())),
                    problem_code=DataAgreementTerminationProblemReportReason.PRINCIPLE_DID_INVALID.value,
                    explain=f"Data agreement principle DID does not match sender DID; Failed to process terminate message for data agreement: {data_agreement_termination_terminate_message.body.data_agreement_id}",
                    data_agreement_id=data_agreement_termination_terminate_message_body.data_agreement_id
                )

                problem_report.assign_thread_id(
                    thid=data_agreement_termination_terminate_message._id
                )

                # Update presentation exchange record with data agreement metadata
                pres_ex_record.data_agreement_problem_report = problem_report.serialize()
                pres_ex_record.data_agreement_status = V10PresentationExchange.DATA_AGREEMENT_PROBLEM_REPORT
                await pres_ex_record.save(context)

                await responder.send_reply(problem_report)

                return

            # Update data agreement event with terminate event
            data_agreement_instance.event.append(
                data_agreement_termination_terminate_message_body.event
            )

            # Update data agreement proof chain with terminate proof
            data_agreement_instance.proof_chain.append(
                data_agreement_termination_terminate_message_body.proof
            )

            # Verify signatures on data agreement
            verkeys = []

            for event in data_agreement_instance.event:
                temp_verkey = DIDMyData.from_did(event.did).public_key_b58
                verkeys.append(temp_verkey)

            valid = await verify_data_agreement(
                data_agreement_instance.serialize(),
                verkeys[-1],
                wallet,
                drop_proof_chain=False
            )

            if not valid:
                self._logger.error(
                    "Data agreement terminate verification failed"
                )

                # Send problem report

                problem_report = DataAgreementTerminationProblemReport(
                    from_did=data_agreement_termination_terminate_message.to_did,
                    to_did=data_agreement_termination_terminate_message.from_did,
                    created_time=str(
                        int(datetime.datetime.utcnow().timestamp())),
                    problem_code=DataAgreementTerminationProblemReportReason.SIGNATURE_VERIFICATION_FAILED.value,
                    explain=f"Data agreement terminate verification failed; Failed to process terminate message for data agreement: {data_agreement_termination_terminate_message.body.data_agreement_id}",
                    data_agreement_id=data_agreement_termination_terminate_message_body.data_agreement_id
                )

                problem_report.assign_thread_id(
                    thid=data_agreement_termination_terminate_message._id
                )

                # Update presentation exchange record with data agreement metadata
                pres_ex_record.data_agreement_problem_report = problem_report.serialize()
                pres_ex_record.data_agreement_status = V10PresentationExchange.DATA_AGREEMENT_PROBLEM_REPORT
                await pres_ex_record.save(context)

                await responder.send_reply(problem_report)

                raise HandlerException(
                    "Data agreement terminate signature verification failed"
                )

            # Update presentation exchange record with data agreement metadata
            pres_ex_record.data_agreement = data_agreement_instance.serialize()
            pres_ex_record.data_agreement_status = V10PresentationExchange.DATA_AGREEMENT_TERMINATE

            await pres_ex_record.save(context)

            # Construct terminate ack message
            data_agreement_terminate_ack = DataAgreementTerminationAck(
                status="TERMINATE OK"
            )

            data_agreement_terminate_ack.assign_thread_id(
                thid=data_agreement_termination_terminate_message._id
            )

            await responder.send_reply(data_agreement_terminate_ack)



from aries_cloudagent.messaging.base_handler import BaseHandler, BaseResponder, RequestContext, HandlerException
from aries_cloudagent.storage.record import StorageRecord
from aries_cloudagent.wallet.base import BaseWallet
from aries_cloudagent.wallet.indy import IndyWallet

from ..messages.data_agreement_reject import DataAgreementNegotiationRejectMessage
from ..manager import ADAManager
from ..models.data_agreement_negotiation_reject_model import DataAgreementNegotiationRejectBody
from ..models.exchange_records.data_agreement_record import DataAgreementV1Record
from ..models.data_agreement_negotiation_offer_model import DataAgreementNegotiationOfferBody, DataAgreementNegotiationOfferBodySchema
from ..models.data_agreement_instance_model import DataAgreementInstance
from ..utils.did.mydata_did import DIDMyData
from ..utils.jsonld.data_agreement import verify_data_agreement_with_proof_chain

from ...patched_protocols.issue_credential.v1_0.models.credential_exchange import (
    V10CredentialExchange
)

from ...patched_protocols.present_proof.v1_0.models.presentation_exchange import (
    V10PresentationExchange
)

import json


class DataAgreementNegotiationRejectMessageHandler(BaseHandler):
    """Handler for data-agreement-negotiation/1.0/reject message."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Message handler logic for data-agreement-negotiation/1.0/reject message."""

        # Assert that the message is of the correct type
        assert isinstance(context.message, DataAgreementNegotiationRejectMessage)

        self._logger.info(
            "Received data-agreement-negotiation/1.0/reject message: \n%s",
            json.dumps(context.message.serialize(), indent=4)
        )

        # Check if connection is ready
        if not context.connection_ready:
            self._logger.info(
                "Connection not active, skipping data-agreement-negotiation/1.0/reject handler: %s",
                context.message_receipt.sender_did,
            )
            return
        
        data_agreement_negotiation_reject_message = context.message
        data_agreement_negotiation_reject_message_body : DataAgreementNegotiationRejectBody = data_agreement_negotiation_reject_message.body

        # Wallet instance from request context
        wallet: IndyWallet = await context.inject(BaseWallet)

        # Initialize ADA manager
        ada_manager = ADAManager(context)

        # Fetch the data agreement instance metadata
        data_agreement_instance_metadata_records = await ada_manager.query_data_agreement_instance_metadata(
            tag_query={
                'data_agreement_id': data_agreement_negotiation_reject_message_body.data_agreement_id,
            }
        )

        # Check if there is a data agreement instance metadata record
        if not data_agreement_instance_metadata_records:
            self._logger.info(
                "Data agreement not found; Failed to handle data agreement reject message for data agreement: %s",
                data_agreement_negotiation_reject_message_body.data_agreement_id,
            )
            return
        
        if len(data_agreement_instance_metadata_records) > 1:
            self._logger.info(
                "Duplicate data agreement records found; Failed to handle data agreement reject message for data agreement: %s",
                data_agreement_negotiation_reject_message_body.data_agreement_id,
            )
            return
        
        data_agreement_instance_metadata_record: StorageRecord = data_agreement_instance_metadata_records[0]

        # Identify the method of use

        if data_agreement_instance_metadata_record.tags.get("method_of_use") == DataAgreementV1Record.METHOD_OF_USE_DATA_SOURCE:
            
            # Fetch exchante record (credential exchange if method of use is "data-source")
            tag_filter = {}
            post_filter = {
                "data_agreement_id": data_agreement_negotiation_reject_message_body.data_agreement_id
            }
            records = await V10CredentialExchange.query(context, tag_filter, post_filter)

            if not records:
                self._logger.info(
                    "Credential exchange record not found; Failed to handle data agreement reject message for data agreement: %s",
                    data_agreement_negotiation_reject_message_body.data_agreement_id,
                )
                return
            
            if len(records) > 1:
                self._logger.info(
                    "Duplicate credential exchange records found; Failed to handle data agreement reject message for data agreement: %s",
                    data_agreement_negotiation_reject_message_body.data_agreement_id,
                )
                return
            
            cred_ex_record: V10CredentialExchange = records[0]

            # Check if the credential exchange record is in the "offer_sent" state
            if cred_ex_record.state != V10CredentialExchange.STATE_OFFER_SENT:
                self._logger.info(
                    "Credential exchange record not in offer sent state; Failed to handle data agreement reject message for data agreement: %s",
                    data_agreement_negotiation_reject_message_body.data_agreement_id,
                )
                return

            # Reconstruct the data agreement with the reject event and proof

            # Deserialise data agreement
            data_agreement_offer: DataAgreementNegotiationOfferBody = DataAgreementNegotiationOfferBodySchema().load(
                cred_ex_record.data_agreement
            )

            # Construct data agreement with proof chain
            data_agreement_with_proof_chain = DataAgreementInstance(
                context=data_agreement_offer.context,
                data_agreement_id=data_agreement_offer.data_agreement_id,
                data_agreement_version=data_agreement_offer.data_agreement_version,
                data_agreement_template_id=data_agreement_offer.data_agreement_template_id,
                data_agreement_template_version=data_agreement_offer.data_agreement_template_version,
                pii_controller_name=data_agreement_offer.pii_controller_name,
                pii_controller_url=data_agreement_offer.pii_controller_url,
                usage_purpose=data_agreement_offer.usage_purpose,
                usage_purpose_description=data_agreement_offer.usage_purpose_description,
                legal_basis=data_agreement_offer.legal_basis,
                method_of_use=data_agreement_offer.method_of_use,
                data_policy=data_agreement_offer.data_policy,
                personal_data=data_agreement_offer.personal_data,
                dpia=data_agreement_offer.dpia,
                event=[
                    data_agreement_offer.event[0],
                    data_agreement_negotiation_reject_message_body.event
                ],
                proof_chain=[
                    data_agreement_offer.proof,
                    data_agreement_negotiation_reject_message_body.proof
                ],
                principle_did=data_agreement_offer.principle_did
            )

            # Principle MyData DID (Data Subject)
            principle_did = DIDMyData.from_did(
                data_agreement_negotiation_reject_message_body.proof.verification_method)

            # Controler MyData DID (Data Controller - Organisation)
            controller_did = DIDMyData.from_did(
                data_agreement_offer.proof.verification_method)

            # Verify signatures on data agreement
            valid = await verify_data_agreement_with_proof_chain(
                data_agreement_with_proof_chain.serialize(),
                [
                    controller_did.public_key_b58,
                    principle_did.public_key_b58
                ],
                wallet
            )

            if not valid:
                self._logger.error(
                    "Data agreement accept verification failed"
                )

                raise HandlerException(
                    "Data agreement accept signature verification failed"
                )

            # Update credential exchange record with data agreement metadata
            cred_ex_record.data_agreement = data_agreement_with_proof_chain.serialize()
            cred_ex_record.data_agreement_status = V10CredentialExchange.DATA_AGREEMENT_REJECT

            await cred_ex_record.save(context)

        if data_agreement_instance_metadata_record.tags.get("method_of_use") == DataAgreementV1Record.METHOD_OF_USE_DATA_USING_SERVICE:
            # Implement data-agreement-negotiation/1.0/reject message handler for data-using-service method-of-use
            # Fetch exchange record (presentation exchange if method of use is "data-using-service")
            tag_filter = {}
            post_filter = {
                "data_agreement_id": data_agreement_negotiation_reject_message_body.data_agreement_id
            }
            records = await V10PresentationExchange.query(context, tag_filter, post_filter)

            if not records:
                self._logger.info(
                    "Presentation exchange record not found; Failed to handle data agreement reject message for data agreement: %s",
                    data_agreement_negotiation_reject_message_body.data_agreement_id,
                )
                return
            
            if len(records) > 1:
                self._logger.info(
                    "Duplicate presentation exchange records found; Failed to handle data agreement reject message for data agreement: %s",
                    data_agreement_negotiation_reject_message_body.data_agreement_id,
                )
                return
            
            pres_ex_record: V10PresentationExchange = records[0]

            # Check if the presentation exchange record is in the "request_sent" state
            if pres_ex_record.state != V10PresentationExchange.STATE_REQUEST_SENT:
                self._logger.info(
                    "Presentation exchange record not in request sent state; Failed to handle data agreement reject message for data agreement: %s",
                    data_agreement_negotiation_reject_message_body.data_agreement_id,
                )
                return

            # Reconstruct the data agreement with the reject event and proof

            # Deserialise data agreement
            data_agreement_offer: DataAgreementNegotiationOfferBody = DataAgreementNegotiationOfferBodySchema().load(
                pres_ex_record.data_agreement
            )

            # Construct data agreement with proof chain
            data_agreement_with_proof_chain = DataAgreementInstance(
                context=data_agreement_offer.context,
                data_agreement_id=data_agreement_offer.data_agreement_id,
                data_agreement_version=data_agreement_offer.data_agreement_version,
                data_agreement_template_id=data_agreement_offer.data_agreement_template_id,
                data_agreement_template_version=data_agreement_offer.data_agreement_template_version,
                pii_controller_name=data_agreement_offer.pii_controller_name,
                pii_controller_url=data_agreement_offer.pii_controller_url,
                usage_purpose=data_agreement_offer.usage_purpose,
                usage_purpose_description=data_agreement_offer.usage_purpose_description,
                legal_basis=data_agreement_offer.legal_basis,
                method_of_use=data_agreement_offer.method_of_use,
                data_policy=data_agreement_offer.data_policy,
                personal_data=data_agreement_offer.personal_data,
                dpia=data_agreement_offer.dpia,
                event=[
                    data_agreement_offer.event[0],
                    data_agreement_negotiation_reject_message_body.event
                ],
                proof_chain=[
                    data_agreement_offer.proof,
                    data_agreement_negotiation_reject_message_body.proof
                ],
                principle_did=data_agreement_offer.principle_did
            )

            # Principle MyData DID (Data Subject)
            principle_did = DIDMyData.from_did(
                data_agreement_negotiation_reject_message_body.proof.verification_method)

            # Controler MyData DID (Data Controller - Organisation)
            controller_did = DIDMyData.from_did(
                data_agreement_offer.proof.verification_method)

            # Verify signatures on data agreement
            valid = await verify_data_agreement_with_proof_chain(
                data_agreement_with_proof_chain.serialize(),
                [
                    controller_did.public_key_b58,
                    principle_did.public_key_b58
                ],
                wallet
            )

            if not valid:
                self._logger.error(
                    "Data agreement accept verification failed"
                )

                raise HandlerException(
                    "Data agreement accept signature verification failed"
                )

            # Update presentation exchange record with data agreement metadata
            pres_ex_record.data_agreement = data_agreement_with_proof_chain.serialize()
            pres_ex_record.data_agreement_status = V10PresentationExchange.DATA_AGREEMENT_REJECT

            await pres_ex_record.save(context)



"""Wallet-related exceptions."""

from aries_cloudagent.core.error import BaseError


class WalletError(BaseError):
    """General wallet exception."""


class WalletNotFoundError(WalletError):
    """Record not found exception."""


class WalletDuplicateError(WalletError):
    """Duplicate record exception."""


class WalletSettingsError(WalletError):
    """Invalid settings exception."""



import json
from aries_cloudagent.messaging.base_handler import BaseHandler, BaseResponder, RequestContext
from dexa_sdk.managers.ada_manager import V2ADAManager
from ..messages.data_agreement_qr_code_initiate import DataAgreementQrCodeInitiateMessage


class DataAgreementQrCodeInitiateHandler(BaseHandler):
    """Handler for qr code initiate message"""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """
        Handler function for qr code initiate message.
        """

        # Assert if received message is of type DataAgreementQrCodeInitiateMessage
        assert isinstance(context.message, DataAgreementQrCodeInitiateMessage)

        self._logger.info(
            "Received data-agreement-qr-code/1.0/initiate message: \n%s",
            json.dumps(context.message.serialize(), indent=4)
        )

        # Initialize ADA manager
        mgr = V2ADAManager(context)

        # Call the function
        await mgr.process_data_agreement_qr_code_initiate_message(
            context.message,
            context.message_receipt,
        )



from aries_cloudagent.messaging.base_handler import BaseHandler, BaseResponder, RequestContext

from ..messages.data_agreement_qr_code_problem_report import DataAgreementQrCodeProblemReport
from ..manager import ADAManager

import json


class DataAgreementQrCodeProblemReportHandler(BaseHandler):
    """Handle for data-agreement-qr-code/1.0/problem-report message"""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """
        Message handler logic for data-agreement-qr-code/1.0/problem-report message.
        """

        # Assert if received message is of type DataAgreementQrCodeProblemReport
        assert isinstance(context.message, DataAgreementQrCodeProblemReport)

        self._logger.info(
            "Received data-agreement-qr-code/1.0/problem-report message: \n%s",
            json.dumps(context.message.serialize(), indent=4)
        )




from aries_cloudagent.messaging.base_handler import BaseHandler, BaseResponder, RequestContext

from ..messages.read_did_response import ReadDIDResponseMessage
from ..manager import ADAManager

import json


class ReadDIDResponseHandler(BaseHandler):
    """
    Read DID response handler class
    """

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """
        Message handler logic
        """

        # Assert if received message is of type ReadDIDResponseMessage
        assert isinstance(context.message, ReadDIDResponseMessage)

        self._logger.info(
            "Received read-did-response message: \n%s",
            json.dumps(context.message.serialize(), indent=4)
        )

        # Check if connection is ready
        if not context.connection_ready:
            self._logger.info(
                "Connection not active, skipping read-did-response handler: %s",
                context.message_receipt.sender_did,
            )
            return
        
        # Initialize ADA manager
        mgr = ADAManager(context)

        # Process the read DID response
        await mgr.process_read_did_response_message(context.message, context.message_receipt)






from aries_cloudagent.messaging.agent_message import AgentMessage, AgentMessageSchema
from marshmallow import EXCLUDE
from ..message_types import DATA_CONTROLLER_DETAILS, PROTOCOL_PACKAGE

# Handler class for data controller details
HANDLER_CLASS = (
    f"{PROTOCOL_PACKAGE}.handlers"
    ".data_controller_details_handler.DataControllerDetailsHandler"
)


class DataControllerDetailsMessage(AgentMessage):
    """
    Message class for data controller details message.
    """

    class Meta:
        # Handler class that can handle this message
        handler_class = HANDLER_CLASS

        # Message type
        message_type = DATA_CONTROLLER_DETAILS

        # Message schema class
        schema_class = "DataControllerDetailsMessageSchema"

    def __init__(self, **kwargs):
        """
        Initialize a DataControllerDetailsMessage message instance.
        """
        super().__init__(**kwargs)


class DataControllerDetailsMessageSchema(AgentMessageSchema):
    """
    Schema class for data controller details message
    """

    class Meta:
        # The message class that this schema is for
        model_class = DataControllerDetailsMessage

        # Unknown fields to exclude from the schema (handled by marshmallow)
        unknown = EXCLUDE



"""DID MyData class and resolver methods."""

from ..wallet.crypto import ed25519_pk_to_curve25519
from ..wallet.key_type import KeyType
from ..wallet.util import b58_to_bytes, bytes_to_b58

from ..diddoc import DIDDoc


class DIDMyData:
    """DID MyData parser and resolver."""

    _key_type: KeyType
    _public_key: bytes

    def __init__(self, public_key: bytes, key_type: KeyType) -> None:
        """Initialize new DIDMyData instance."""
        self._public_key = public_key
        self._key_type = key_type

    @classmethod
    def from_public_key(cls, public_key: bytes, key_type: KeyType) -> "DIDMyData":
        """Initialize new DIDMyData instance from public key and key type."""

        return cls(public_key, key_type)

    @classmethod
    def from_public_key_b58(cls, public_key: str, key_type: KeyType) -> "DIDMyData":
        """Initialize new DIDMyData instance from base58 encoded public key and key type."""
        public_key_bytes = b58_to_bytes(public_key)
        return cls.from_public_key(public_key_bytes, key_type)

    @classmethod
    def from_fingerprint(cls, fingerprint: str) -> "DIDMyData":
        """Initialize new DIDMyData instance from multibase encoded fingerprint.

        The fingerprint contains both the public key and key type.
        """
        # Assert fingerprint is in multibase format
        assert fingerprint[0] == "z"

        # Get key bytes, remove multicodec prefix
        key_bytes_with_prefix = b58_to_bytes(fingerprint[1:])

        # Get associated key type with prefixed bytes
        key_type = KeyType.from_prefixed_bytes(key_bytes_with_prefix)

        if not key_type:
            raise Exception(
                f"No key type for prefixed public key '{key_bytes_with_prefix}' found."
            )

        # Remove the prefix bytes to get the public key
        prefix_len = len(key_type.multicodec_prefix)
        public_key_bytes = key_bytes_with_prefix[prefix_len:]

        return cls(public_key_bytes, key_type)

    @classmethod
    def from_did(cls, did: str) -> "DIDMyData":
        """Initialize a new DIDMyData instance from a fully qualified did:mydata string.

        Extracts the fingerprint from the did:mydata and uses that to constrcut the did:mydata.
        """
        did_parts = did.split("#")
        _, fingerprint = did_parts[0].split("did:mydata:")

        return cls.from_fingerprint(fingerprint)

    @property
    def prefixed_public_key(self) -> bytes:
        """Getter for multicodec prefixed public key."""
        return b"".join([self.key_type.multicodec_prefix, self.public_key])

    @property
    def fingerprint(self) -> str:
        """Getter for DID MyData fingerprint."""
        return f"z{bytes_to_b58(self.prefixed_public_key)}"

    @property
    def did(self) -> str:
        """Getter for full did:mydata string."""
        return f"did:mydata:{self.fingerprint}"

    @property
    def did_doc(self) -> dict:
        """Getter for did document associated with did:mydata."""
        resolver = DID_KEY_RESOLVERS[self.key_type]
        return resolver(self)

    @property
    def public_key(self) -> bytes:
        """Getter for public key."""
        return self._public_key

    @property
    def public_key_b58(self) -> str:
        """Getter for base58 encoded public key."""
        return bytes_to_b58(self.public_key)

    @property
    def key_type(self) -> KeyType:
        """Getter for key type."""
        return self._key_type

    @property
    def key_id(self) -> str:
        """Getter for key id."""
        return f"{self.did}#{self.fingerprint}"

def construct_did_key_ed25519(did_key: "DIDMyData") -> dict:
    """Construct Ed25519 did:mydata.

    Args:
        did_key (DIDMyData): DID MyData instance to parse ed25519 did:mydata document from

    Returns:
        dict: The ed25519 did:mydata did document

    """
    curve25519 = ed25519_pk_to_curve25519(did_key.public_key)
    x25519 = DIDMyData.from_public_key(curve25519, KeyType.X25519)

    did_doc = construct_did_signature_key_base(
        id=did_key.did,
        key_id=did_key.key_id,
        verification_method={
            "id": did_key.key_id,
            "type": "Ed25519VerificationKey2018",
            "controller": did_key.did,
            "publicKeyBase58": did_key.public_key_b58,
        },
    )

    # Ed25519 has pair with X25519
    did_doc["keyAgreement"].append(
        {
            "id": f"{did_key.did}#{x25519.fingerprint}",
            "type": "X25519KeyAgreementKey2019",
            "controller": did_key.did,
            "publicKeyBase58": bytes_to_b58(curve25519),
        }
    )

    return did_doc


def construct_did_signature_key_base(
    *, id: str, key_id: str, verification_method: dict
):
    """Create base DID MyData structure to use for most signature keys.

    May not be suitable for all DID MyData types

    """

    return {
        "@context": DIDDoc.CONTEXT,
        "id": id,
        "verificationMethod": [verification_method],
        "authentication": [key_id],
        "assertionMethod": [key_id],
        "capabilityDelegation": [key_id],
        "capabilityInvocation": [key_id],
        "keyAgreement": [],
    }


DID_KEY_RESOLVERS = {
    KeyType.ED25519: construct_did_key_ed25519,
}



"""Presentation ack message handler."""
import json

from aries_cloudagent.messaging.base_handler import (
    BaseHandler,
    BaseResponder,
    HandlerException,
    RequestContext,
)

from ..messages.data_agreement_terminate_ack import DataAgreementTerminationAck


class DataAgreementTerminationAckHandler(BaseHandler):
    """Message handler class for data agreement terminate acks."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """
        Message handler logic for data agreement terminate acks.

        Args:
            context: request context
            responder: responder callback
        """

        self._logger.debug("DataAgreementTerminationAckHandler called with context %s", context)

        assert isinstance(context.message, DataAgreementTerminationAck)

        self._logger.info(
            "Received data agreement terminate ack message: %s",
            json.dumps(context.message.serialize(), indent=4),
        )



from aries_cloudagent.messaging.base_handler import BaseHandler, BaseResponder, RequestContext

from ..messages.json_ld_processed import JSONLDProcessedMessage
from ..manager import ADAManager

import json


class JSONLDProcessedHandler(BaseHandler):
    """Handle for json-ld/1.0/processed-data message"""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """
        Message handler logic for json-ld/1.0/processed-data message.
        """

        # Assert if received message is of type JSONLDProcessedMessage
        assert isinstance(context.message, JSONLDProcessedMessage)

        self._logger.info(
            "Received json-ld/1.0/processed-data message: \n%s",
            json.dumps(context.message.serialize(), indent=4)
        )

        # Initialize ADA manager
        ada_manager = ADAManager(context)

        # Call the function

        await ada_manager.process_json_ld_processed_message(
            json_ld_processed_message=context.message,
            receipt=context.message_receipt,
        )




from aries_cloudagent.messaging.base_handler import BaseHandler, BaseResponder, RequestContext
from dexa_sdk.managers.ada_manager import V2ADAManager
from ..messages.existing_connections import ExistingConnectionsMessage


class ExistingConnectionsMessageHandler(BaseHandler):
    """Handler for existing connections message"""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """
        Handle function.
        """

        # Assert message type.
        assert isinstance(context.message, ExistingConnectionsMessage)

        # Initialize ADA manager
        mgr = V2ADAManager(context)

        # Call the function
        await mgr.process_existing_connections_message(
            context.message,
            context.message_receipt,
        )



"""Wallet utility functions."""

import base58
import base64

import nacl.utils
import nacl.bindings


def random_seed() -> bytes:
    """
    Generate a random seed value.

    Returns:
        A new random seed

    """
    return nacl.utils.random(nacl.bindings.crypto_box_SEEDBYTES)


def pad(val: str) -> str:
    """Pad base64 values if need be: JWT calls to omit trailing padding."""
    padlen = 4 - len(val) % 4
    return val if padlen > 2 else (val + "=" * padlen)


def unpad(val: str) -> str:
    """Remove padding from base64 values if need be."""
    return val.rstrip("=")


def b64_to_bytes(val: str, urlsafe=False) -> bytes:
    """Convert a base 64 string to bytes."""
    if urlsafe:
        return base64.urlsafe_b64decode(pad(val))
    return base64.b64decode(pad(val))


def b64_to_str(val: str, urlsafe=False, encoding=None) -> str:
    """Convert a base 64 string to string on input encoding (default utf-8)."""
    return b64_to_bytes(val, urlsafe).decode(encoding or "utf-8")


def bytes_to_b64(val: bytes, urlsafe=False, pad=True, encoding: str = "ascii") -> str:
    """Convert a byte string to base 64."""
    b64 = (
        base64.urlsafe_b64encode(val).decode(encoding)
        if urlsafe
        else base64.b64encode(val).decode(encoding)
    )
    return b64 if pad else unpad(b64)


def str_to_b64(val: str, urlsafe=False, encoding=None, pad=True) -> str:
    """Convert a string to base64 string on input encoding (default utf-8)."""
    return bytes_to_b64(val.encode(encoding or "utf-8"), urlsafe, pad)


def set_urlsafe_b64(val: str, urlsafe: bool = True) -> str:
    """Set URL safety in base64 encoding."""
    if urlsafe:
        return val.replace("+", "-").replace("/", "_")
    return val.replace("-", "+").replace("_", "/")


def b58_to_bytes(val: str) -> bytes:
    """Convert a base 58 string to bytes."""
    return base58.b58decode(val)


def bytes_to_b58(val: bytes) -> str:
    """Convert a byte string to base 58."""
    return base58.b58encode(val).decode("ascii")


def full_verkey(did: str, abbr_verkey: str) -> str:
    """Given a DID and abbreviated verkey, return the full verkey."""
    return (
        bytes_to_b58(b58_to_bytes(did.split(":")[-1]) + b58_to_bytes(abbr_verkey[1:]))
        if abbr_verkey.startswith("~")
        else abbr_verkey
    )


def abbr_verkey(full_verkey: str, did: str = None) -> str:
    """Given a full verkey and DID, return the abbreviated verkey."""
    did_len = len(b58_to_bytes(did.split(":")[-1])) if did else 16
    return f"~{bytes_to_b58(b58_to_bytes(full_verkey)[did_len:])}"



import typing

from aries_cloudagent.messaging.models.base import BaseModel, BaseModelSchema
from aries_cloudagent.messaging.valid import UUIDFour

from marshmallow import fields, EXCLUDE
from marshmallow.exceptions import ValidationError

from ..utils.diddoc import DIDDoc
from ..utils.regex import MYDATA_DID, MyDataDID
from ..utils.verification_method import PublicKeyType


class DIDDocWrapper(fields.Field):
    """Field that loads and serializes DIDDoc."""

    def _serialize(self, value, attr, obj, **kwargs):
        """
        Serialize the DIDDoc.

        Args:
            value: The value to serialize

        Returns:
            The serialized DIDDoc

        """
        return value.serialize()

    def _deserialize(self, value, attr, data, **kwargs):
        """
        Deserialize a value into a DIDDoc.

        Args:
            value: The value to deserialize

        Returns:
            The deserialized value

        """
        return DIDDoc.deserialize(value)

    def _validate(self, value: DIDDoc):
        if not value.validate():
            raise ValidationError("MyData DIDDoc is not valid.")


class MyDataDIDBody(BaseModel):
    class Meta:
        schema_class = "MyDataDIDBodySchema"

    def __init__(self, *, did_doc: DIDDoc, **kwargs):
        super().__init__(**kwargs)
        self.did_doc = did_doc


class MyDataDIDBodySchema(BaseModelSchema):
    class Meta:
        model_class = MyDataDIDBody
        unknown = EXCLUDE

    did_doc = DIDDocWrapper(
        data_key="did",
        required=False
    )


class MyDataDIDDocService(BaseModel):
    """
    Service information for a DID Document.
    """
    class Meta:
        # Schema class
        schema_class = "MyDataDIDDocServiceSchema"

    def __init__(
        self,
        *,
        service_id: str = None,
        service_type: str = None,
        service_priority: int = 0,
        recipient_keys: typing.List[str] = None,
        service_endpoint: str = None,
        **kwargs
    ):
        """
        Initialize a DID Document Service object.

        Args:
            service_id: The service ID
            service_type: The service type
            service_priority: The service priority
            recipient_keys: The recipient keys
            service_endpoint: The service endpoint
        """

        super().__init__(**kwargs)

        # Service ID
        self.service_id = service_id

        # Service type
        self.service_type = service_type

        # Service priority
        self.service_priority = service_priority

        # Recipient keys
        self.recipient_keys = recipient_keys

        # Service endpoint
        self.service_endpoint = service_endpoint


class MyDataDIDDocServiceSchema(BaseModelSchema):
    """
    Schema for DID Document Service. 
    """
    class Meta:

        # Model class
        model_class = MyDataDIDDocService

        # Unknown fields are excluded.
        unknown = EXCLUDE

    # Service ID
    service_id = fields.Str(data_key="id", required=True, example=f"did:mydata:{MyDataDID.EXAMPLE};didcomm")

    # Service type
    service_type = fields.Str(data_key="type", required=True, example="DIDComm")

    # Service priority
    service_priority = fields.Int(data_key="priority", required=True, example=1)

    # Recipient keys
    recipient_keys = fields.List(
        fields.Str(required=True, example=MyDataDID.EXAMPLE),
        data_key="recipientKeys",
        required=True
    )

    # Service endpoint
    service_endpoint = fields.Str(data_key="serviceEndpoint", required=True, example="https://didcomm.org")


class MyDataDIDDocAuthentication(BaseModel):
    """
    Authentication information for a DID Document.
    """
    class Meta:
        # Schema class
        schema_class = "MyDataDIDDocAuthenticationSchema"

    def __init__(self, *, authentication_type: str = None, public_key: str = None, **kwargs):
        """
        Initialize a DID Document Authentication object.

        Args:
            authentication_type: The authentication type
            public_key: The public key
        """
        super().__init__(**kwargs)

        # Set attributes
        self.authentication_type = authentication_type
        self.public_key = public_key


class MyDataDIDDocAuthenticationSchema(BaseModelSchema):
    """
    Schema for DID Document Authentication.

    """
    class Meta:

        # Model class
        model_class = MyDataDIDDocAuthentication

        # Unknown fields are excluded.
        unknown = EXCLUDE

    # The authentication type
    authentication_type = fields.Str(data_key="type", required=True, example=PublicKeyType.ED25519_SIG_2018.authn_type)

    # The public key
    public_key = fields.Str(data_key="publicKey", required=True, example=f"did:mydata:f{MyDataDID.EXAMPLE}#1")


class MyDataDIDDocVerificationMethod(BaseModel):
    """
    A DID Document Verification Method.
    """

    class Meta:
        # Schema class
        schema_class = "MyDataDIDDocVerificationMethodSchema"

    def __init__(
        self,
        *,
        verification_method_id: str = None,
        verification_method_type: str = None,
        controller: str = None,
        public_key_base58: str = None,
        **kwargs
    ):
        """
        Initialize a DID Document Verification Method.

        Args:
            verification_method_id: The verification method id
            verification_method_type: The verification method type
            controller: The controller
            public_key_base58: The public key base58
        """

        # Initialize the super class
        super().__init__(**kwargs)

        # Set attributes
        self.verification_method_id = verification_method_id
        self.verification_method_type = verification_method_type
        self.controller = controller
        self.public_key_base58 = public_key_base58


class MyDataDIDDocVerificationMethodSchema(BaseModelSchema):
    """
    A DID Document Verification Method Schema.
    """

    class Meta:
        # Schema class
        model_class = MyDataDIDDocVerificationMethod

    # Verification method id
    verification_method_id = fields.Str(
        data_key="id",
        required=True,
        example=f"did:mydata:{MyDataDID.EXAMPLE}#1"
    )

    # Verification method type
    verification_method_type = fields.Str(
        data_key="type",
        required=True,
        example=PublicKeyType.ED25519_SIG_2018.ver_type
    )

    # Controller
    controller = fields.Str(
        data_key="controller",
        required=True,
        example=f"did:mydata:{MyDataDID.EXAMPLE}"
    )

    # Public key base58
    public_key_base58 = fields.Str(
        data_key="publicKeyBase58",
        required=True,
        example=f"{MyDataDID.EXAMPLE}"
    )


class MyDataDIDDoc(BaseModel):
    """
    MyData DIDDoc model
    """
    class Meta:
        # Schema class
        schema_class = "MyDataDIDDocSchema"

    def __init__(self,
                 *,
                 context: str = None,
                 diddoc_id: str = None,
                 verification_method: typing.List[MyDataDIDDocVerificationMethod] = None,
                 authentication: typing.List[MyDataDIDDocAuthentication] = None,
                 service: typing.List[MyDataDIDDocService] = None,
                 **kwargs):
        """
        Initialize a MyData DIDDoc model.

        Args:
            context: The DIDDoc context
            diddoc_id: The DIDDoc id
            verification_method: The verification method
            authentication: The authentication method
            service: The service
            kwargs: The extra arguments

        """

        # Call parent constructor
        super().__init__(**kwargs)

        # Set attributes
        self.context = context
        self.diddoc_id = diddoc_id
        self.verification_method = verification_method
        self.authentication = authentication
        self.service = service


class MyDataDIDDocSchema(BaseModelSchema):
    """
    MyData DIDDoc schema
    """
    class Meta:
        # Model class
        model_class = MyDataDIDDoc

        # Unknown fields are excluded
        unknown = EXCLUDE

    # The DIDDoc context
    context = fields.Str(
        required=True,
        data_key="@context",
        example=DIDDoc.CONTEXT,
        description="The DIDDoc context"
    )

    # The DIDDoc id
    diddoc_id = fields.Str(
        required=True,
        data_key="id",
        example=f"did:mydata:{MyDataDID.EXAMPLE}",
    )

    # The verification method
    verification_method = fields.List(
        fields.Nested(MyDataDIDDocVerificationMethodSchema),
        required=True
    )

    # The authentication method
    authentication = fields.List(
        fields.Nested(MyDataDIDDocAuthenticationSchema),
        required=True
    )

    # The service
    service = fields.List(
        fields.Nested(MyDataDIDDocServiceSchema),
        required=True
    )


class MyDataDIDResponseBody(BaseModel):
    """
    MyData DID response body model
    """
    class Meta:

        # Schema class
        schema_class = "MyDataDIDResponseBodySchema"

    def __init__(self, *, did_doc: MyDataDIDDoc = None, version: str = None, status: str = None, **kwargs):
        """
        Initialize a MyData DID response body model.

        Args:
            did_doc: The DIDDoc
            version: The version
        """
        super().__init__(**kwargs)

        # Set attributes
        self.did_doc = did_doc
        self.version = version
        self.status = status


class MyDataDIDResponseBodySchema(BaseModelSchema):
    """
    MyData DID response body schema
    """
    class Meta:

        # Model class
        model_class = MyDataDIDResponseBody

        # Unknown fields are excluded
        unknown = EXCLUDE
    
    # The DIDDoc
    did_doc = fields.Nested(MyDataDIDDocSchema, required=True)

    # The version
    version = fields.Str(data_key="version")

    # The status
    status = fields.Str(data_key="status")


from aries_cloudagent.messaging.models.base import BaseModel, BaseModelSchema
from marshmallow import fields, EXCLUDE


class DataController(BaseModel):
    """
    Data controller model class
    """
    class Meta:
        schema_class = "DataControllerSchema"

    def __init__(self,
                 *,
                 organisation_did: str = None,
                 organisation_name: str = None,
                 cover_image_url: str = None,
                 logo_image_url: str = None,
                 location: str = None,
                 organisation_type: str = None,
                 description: str = None,
                 policy_url: str = None,
                 eula_url: str = None,
                 **kwargs):
        super().__init__(**kwargs)

        self.organisation_did = organisation_did
        self.organisation_name = organisation_name
        self.cover_image_url = cover_image_url
        self.logo_image_url = logo_image_url
        self.location = location
        self.organisation_type = organisation_type
        self.description = description
        self.policy_url = policy_url
        self.eula_url = eula_url


class DataControllerSchema(BaseModelSchema):
    """
    Data controller schema class
    """
    class Meta:
        model_class = DataController
        unknown = EXCLUDE

    organisation_did = fields.Str()
    organisation_name = fields.Str()
    cover_image_url = fields.Str()
    logo_image_url = fields.Str()
    location = fields.Str()
    organisation_type = fields.Str()
    description = fields.Str()
    policy_url = fields.Str()
    eula_url = fields.Str()



from aiohttp import web
from ..data_agreement_auditor_functions_routes import (
    query_data_agreement_instances
)
from ..data_agreement_core_functions_routes import (
    create_and_store_data_agreement_in_wallet_v2,
    publish_data_agreement_handler,
    query_data_agreements_in_wallet,
    update_data_agreement_in_wallet_v2,
    delete_data_agreement_in_wallet,
    query_da_personal_data_in_wallet,
    update_da_personal_data_in_wallet,
    delete_da_personal_data_in_wallet,
    generate_data_agreement_qr_code_payload,
    query_data_agreement_qr_code_metadata_records_handler,
    remove_data_agreement_qr_code_metadata_record_handler,
    send_data_agreements_qr_code_workflow_initiate_handler,
    fetch_customer_identification_da_handler,
    configure_customer_identification_da_handler
)

from ..data_controller_functions_routes import (
    send_data_controller_details_message_handler,
    update_data_controller_details
)

from ..jsonld_routes import (
    send_json_ld_didcomm_processed_data_message_handler,
)

from ..mydata_did_operations_routes import (
    send_read_did_message_to_mydata_did_registry,
    mydata_did_remote_records_list,
)


from ..connection_routes import (
    wellknown_connection_handler,
    v2_connections_create_invitation,
    send_existing_connections_message_handler,
    get_existing_connections_handler,
    connections_list_v2,
    generate_firebase_dynamic_link_for_connection_invitation_handler
)

# Routes defined in ADA
ROUTES_ADA = [
    web.post(
        "/v1/mydata-did/didcomm/read-did/{did}",
        send_read_did_message_to_mydata_did_registry,
    ),
    web.get(
        "/v1/mydata-did/remote",
        mydata_did_remote_records_list,
        allow_head=False,
    ),
    web.post(
        "/v1/data-agreements",
        create_and_store_data_agreement_in_wallet_v2,
    ),
    web.post(
        "/v1/data-agreements/{template_id}/publish",
        publish_data_agreement_handler,
    ),
    web.get(
        "/v1/data-agreements", query_data_agreements_in_wallet, allow_head=False
    ),
    web.put(
        "/v1/data-agreements/{template_id}",
        update_data_agreement_in_wallet_v2,
    ),
    web.delete(
        "/v1/data-agreements/{template_id}",
        delete_data_agreement_in_wallet,
    ),
    web.get(
        "/v1/data-agreements/personal-data",
        query_da_personal_data_in_wallet,
        allow_head=False,
    ),
    web.put(
        "/v1/data-agreements/personal-data/{attribute_id}",
        update_da_personal_data_in_wallet,
    ),
    web.delete(
        "/v1/data-agreements/personal-data/{attribute_id}",
        delete_da_personal_data_in_wallet,
    ),
    web.get(
        "/v1/auditor/data-agreements/instances",
        query_data_agreement_instances,
        allow_head=False,
    ),
    web.post(
        "/v1/data-agreements/{template_id}/customer-identification",
        configure_customer_identification_da_handler
    ),
    web.get(
        "/v1/data-agreements/customer-identification",
        fetch_customer_identification_da_handler,
        allow_head=False,
    ),
    web.get(
        "/v1/.well-known/did-configuration.json",
        wellknown_connection_handler,
        allow_head=False,
    ),
    web.post(
        "/v1/data-agreements/{template_id}/qr",
        generate_data_agreement_qr_code_payload,
    ),
    web.get(
        "/v1/data-agreements/{template_id}/qr",
        query_data_agreement_qr_code_metadata_records_handler,
        allow_head=False,
    ),
    web.delete(
        "/v1/data-agreements/{template_id}/qr/{qr_id}",
        remove_data_agreement_qr_code_metadata_record_handler,
    ),
    web.post(
        "/v1/data-agreements/qr/{qr_id}/workflow-initiate/connections/{connection_id}",
        send_data_agreements_qr_code_workflow_initiate_handler,
    ),
    web.post(
        "/v1/json-ld/didcomm/processed-data/connections/{connection_id}",
        send_json_ld_didcomm_processed_data_message_handler,
    ),
    web.post(
        "/v2/connections/create-invitation", v2_connections_create_invitation
    ),
    web.post(
        "/v1/connections/{conn_id}/invitation/firebase",
        generate_firebase_dynamic_link_for_connection_invitation_handler,
    ),
    web.post(
        "/v1/data-controller/didcomm/details/connections/{connection_id}",
        send_data_controller_details_message_handler,
    ),
    web.post(
        "/v1/connections/{conn_id}/existing",
        send_existing_connections_message_handler,
    ),
    web.get(
        "/v1/connections/{conn_id}/existing",
        get_existing_connections_handler,
        allow_head=False,
    ),
    web.get(
        "/v2/connections",
        connections_list_v2,
        allow_head=False
    ),
    web.post(
        "/v1/data-controller",
        update_data_controller_details
    )
]



from os import environ
import typing
from typing import Any

from marshmallow import fields, validate

from aries_cloudagent.messaging.models.base_record import BaseExchangeRecord, BaseExchangeSchema
from aries_cloudagent.messaging.valid import UUIDFour

from ..data_agreement_model import DataAgreementPersonalDataRestriction, DataAgreementPersonalDataRestrictionSchema

class DataAgreementPersonalDataRecord(BaseExchangeRecord):
    """
    DataAgreementPersonalDataRecord model class for serialisation/deserialisation of data agreement personal data records to/from wallet.
    """
    class Meta:
        # Data Agreement Personal Data Record schema class
        schema_class = "DataAgreementPersonalDataRecordSchema"

    # Wallet record type
    RECORD_TYPE = "data_agreement_personal_data_record"

    # Wallet record identifier field
    RECORD_ID_NAME = "personal_data_id"

    # Webhook topic name for this record type
    WEBHOOK_TOPIC = None

    # Wallet record tags used for filtering
    # Note: These are not tags for the ledger, but rather tags for the wallet
    #      to group records.
    TAG_NAMES = {"~attribute_category", "~attribute_sensitive", "~da_template_id"}

    def __init__(
        self,
        *,
        personal_data_id: str = None,
        attribute_name: str = None,
        attribute_category: str = "Other",
        attribute_sensitive: str = "true",
        attribute_description: str = "Nil",
        state: str = None,
        restrictions: typing.List[dict] = None,
        da_template_id: str = None,
        da_template_version: int = None,
        **kwargs
    ):
        """
        Initialise a new DataAgreementPersonalDataRecord instance.

        Args:
            personal_data_id: The unique identifier for the data agreement personal data record.
            attribute_name: The name of the attribute.
            attribute_category: The category of the attribute.
            attribute_sensitive: The sensitive flag of the attribute.
            state: The state of the data agreement personal data record.
        """
        super().__init__(personal_data_id, state, **kwargs)
        self.state = state
        self.attribute_name = attribute_name
        self.attribute_category = attribute_category
        self.attribute_sensitive = attribute_sensitive
        self.attribute_description = attribute_description
        self.restrictions = restrictions
        self.da_template_id = da_template_id
        self.da_template_version = da_template_version

    @property
    def personal_data_id(self) -> str:
        """
        Get the data agreement personal data record identifier.

        Returns:
            The data agreement personal data record identifier.
        """
        return self._id

    @property
    def record_value(self) -> dict:
        """Accessor for JSON record value generated for this transaction record."""
        return {
            prop: getattr(self, prop)
            for prop in (
                "attribute_name",
                "attribute_category",
                "attribute_sensitive",
                "attribute_description",
                "restrictions",
                "da_template_id",
                "da_template_version"
            )
        }

    def __eq__(self, other: Any) -> bool:
        """Comparison between records."""
        return super().__eq__(other)


class DataAgreementPersonalDataRecordSchema(BaseExchangeSchema):

    class Meta:
        """DataAgreementPersonalDataRecordSchema metadata."""

        # Model class
        model_class = DataAgreementPersonalDataRecord
    

    personal_data_id = fields.Str(
        required=True,
        example=UUIDFour.EXAMPLE
    )

    attribute_name = fields.Str(
        required=True,
        description="The name of the attribute.",
        example="name"
    )

    attribute_description = fields.Str(
        required=True,
        description="The description of the attribute.",
        example="Name of the customer"
    )

    attribute_category = fields.Str(
        required=True,
        description="The category of the attribute.",
        example="personal"
    )


    attribute_sensitive = fields.Str(
        required=True,
        description="The sensitive flag of the attribute.",
        example="true"
    )

    restrictions = fields.List(fields.Nested(DataAgreementPersonalDataRestrictionSchema), required=False)

    da_template_id = fields.Str(
        required=True,
        description="The data agreement template identifier.",
        example=UUIDFour.EXAMPLE
    )

    da_template_version = fields.Int(
        example=1,
        description="Data agreement template version"
    )





import typing
import datetime

from marshmallow import EXCLUDE, fields, validate, validates
from marshmallow.exceptions import ValidationError
from aries_cloudagent.messaging.models.base import BaseModel, BaseModelSchema
from aries_cloudagent.messaging.valid import UUIDFour

from ..models.data_agreement_model import (
    DataAgreementDataPolicy,
    DataAgreementDataPolicySchema,
    DataAgreementPersonalData,
    DataAgreementPersonalDataSchema,
    DataAgreementDPIA,
    DataAgreementDPIASchema,
    DATA_AGREEMENT_V1_SCHEMA_CONTEXT,
)
from ..utils.jsonld import ED25519_2018_CONTEXT_URL
from ..utils.util import current_datetime_in_iso8601
from ..utils.regex import MYDATA_DID


class DataAgreementProof(BaseModel):
    """
    Data agreement proof model class
    """
    class Meta:
        # Schema class
        schema_class = "DataAgreementProofSchema"

        # Unknown fields are excluded
        unknown = EXCLUDE

    def __init__(
        self,
        *,
        proof_id: str = None,
        proof_type: str = None,
        created: str = None,
        verification_method: str = None,
        proof_purpose: str = None,
        proof_value: str = None,
        **kwargs
    ):
        # Call parent constructor
        super().__init__(**kwargs)

        # Set attributes
        self.proof_id = proof_id
        self.proof_type = proof_type
        self.created = created
        self.verification_method = verification_method
        self.proof_purpose = proof_purpose
        self.proof_value = proof_value


class DataAgreementProofSchema(BaseModelSchema):
    """
    Data agreement proof schema class
    """
    class Meta:
        # Model class
        model_class = DataAgreementProof

    # Proof identifier
    proof_id = fields.Str(
        data_key="id",
        example="did:mydata:123456789abcdefghi#1",
        description="Proof identifier",
        required=False
    )

    # Proof type
    proof_type = fields.Str(
        data_key="type",
        example="Ed25519Signature2018",
        description="Proof type",
        required=False
    )

    # Created
    created = fields.Str(
        data_key="created",
        example=current_datetime_in_iso8601(),
        description="Proof created date time in ISO 8601 format",
        required=False
    )

    # Verification method
    verification_method = fields.Str(
        data_key="verificationMethod",
        example="did:mydata:123456789abcdefghi",
        description="Verification method",
        required=False
    )

    # Proof purpose
    proof_purpose = fields.Str(
        data_key="proofPurpose",
        example="contractAgreement",
        description="Proof purpose",
        required=False
    )

    # Proof value
    proof_value = fields.Str(
        data_key="proofValue",
        example="123456789abcdefghi",
        description="Proof value",
        required=False
    )


class DataAgreementDummy(BaseModel):
    """
    Data agreement dummy model class
    """
    class Meta:
        # Schema class
        schema_class = "DataAgreementDummySchema"

        # Unknown fields are excluded
        unknown = EXCLUDE

    def __init__(
        self,
        *,
        dummy_id: str = None,
        **kwargs
    ):
        # Call parent constructor
        super().__init__(**kwargs)

        # Set attributes
        self.dummy_id = dummy_id

class DataAgreementDummySchema(BaseModelSchema):
    """
    Data agreement dummy schema class
    """
    class Meta:
        # Model class
        model_class = DataAgreementDummy

    # Dummy identifier
    dummy_id = fields.Str(
        data_key="id",
        example="did:mydata:123456789abcdefghi#1",
        description="Dummy identifier",
        required=False
    )


class DataAgreementEvent(BaseModel):
    """
    Data agreement event model class
    """

    # States
    STATE_ACCEPT = "accept"
    STATE_REJECT = "reject"
    STATE_OFFER = "offer"
    STATE_TERMINATE = "terminate"

    class Meta:
        # Schema class
        schema_class = "DataAgreementEventSchema"

        # Unknown fields are excluded
        unknown = EXCLUDE

    def __init__(
        self,
        *,
        event_id: str = None,
        time_stamp: str = None,
        did: str = None,
        state: str = None,
        **kwargs
    ):
        # Call parent constructor
        super().__init__(**kwargs)

        # Set attributes
        self.event_id = event_id
        self.time_stamp = time_stamp
        self.did = did
        self.state = state


class DataAgreementEventSchema(BaseModelSchema):
    """
    Data agreement event schema class
    """
    class Meta:
        # Model class
        model_class = DataAgreementEvent

    event_id = fields.Str(
        data_key="id",
        example="did:mydata:123456789abcdefghi#1",
        description="Data agreement event identifier",
    )

    # Time stamp
    time_stamp = fields.Str(
        data_key="time_stamp",
        example=str(datetime.datetime.utcnow().replace(
            tzinfo=datetime.timezone.utc).isoformat()),
        description="Data agreement event timestamp in ISO 8601 UTC date time format"
    )

    # Event origin DID
    did = fields.Str(
        data_key="did",
        description="MyData decentralised identifier",
        **MYDATA_DID
    )

    # State
    state = fields.Str(
        description="State of the event",
        example="capture"
    )


class DataAgreementNegotiationOfferBody(BaseModel):
    """Data Agreement Negotiation Offer Body"""

    class Meta:
        """Data Agreement Negotiation Offer Body metadata"""

        schema_class = "DataAgreementNegotiationOfferBodySchema"

    def __init__(
        self,
        *,
        context: typing.List[str],
        data_agreement_id: str = None,
        data_agreement_version: int = None,
        data_agreement_template_id: str = None,
        data_agreement_template_version: int = None,
        pii_controller_name: str = None,
        pii_controller_url: str = None,
        usage_purpose: str = None,
        usage_purpose_description: str = None,
        legal_basis: str = None,
        method_of_use: str = None,
        data_policy: DataAgreementDataPolicy = None,
        personal_data: typing.List[DataAgreementPersonalData] = None,
        dpia: DataAgreementDPIA = None,
        proof: DataAgreementProof = None,
        event: typing.List[DataAgreementEvent] = None,
        principle_did: str = None,
        **kwargs
    ):
        """Data Agreement Negotiation Offer Body init"""

        super().__init__(**kwargs)

        # Set attributes
        self.context = context
        self.data_agreement_id = data_agreement_id
        self.data_agreement_version = data_agreement_version
        self.data_agreement_template_id = data_agreement_template_id
        self.data_agreement_template_version = data_agreement_template_version
        self.pii_controller_name = pii_controller_name
        self.pii_controller_url = pii_controller_url
        self.usage_purpose = usage_purpose
        self.usage_purpose_description = usage_purpose_description
        self.legal_basis = legal_basis
        self.method_of_use = method_of_use
        self.data_policy = data_policy
        self.personal_data = personal_data
        self.dpia = dpia
        self.proof = proof
        self.event = event
        self.principle_did = principle_did
        


class DataAgreementNegotiationOfferBodySchema(BaseModelSchema):
    """Data Agreement Negotiation Offer Body schema"""

    class Meta:
        """Data Agreement Negotiation Offer Body schema metadata"""

        model_class = DataAgreementNegotiationOfferBody

    # Context
    context = fields.List(
        fields.Str(),
        example=[DATA_AGREEMENT_V1_SCHEMA_CONTEXT, ED25519_2018_CONTEXT_URL],
        data_key="@context",
        description="Context",
    )

    # Data agreement id
    data_agreement_id = fields.Str(
        data_key="id",
        example=UUIDFour.EXAMPLE,
        description="Data agreement identifier"
    )

    # Data agreement version
    data_agreement_version = fields.Int(
        data_key="version",
        example=1,
        description="Data agreement version"
    )

    # Data agreement template id
    data_agreement_template_id = fields.Str(
        data_key="template_id",
        example=UUIDFour.EXAMPLE,
        description="Data agreement template identifier"
    )

    # Data agreement template version
    data_agreement_template_version = fields.Int(
        data_key="template_version",
        example=1,
        description="Data agreement template version"
    )

    # Data agreement data controller name
    # i.e. Organization name of the data controller
    pii_controller_name = fields.Str(
        data_key="data_controller_name",
        example="Happy Shopping AB",
        description="PII controller name"
    )

    # Data agreement data controller URL
    pii_controller_url = fields.Str(
        data_key="data_controller_url",
        example="https://www.happyshopping.com",
        description="PII controller URL"
    )

    # Data agreement usage purpose
    usage_purpose = fields.Str(
        data_key="purpose",
        example="Customized shopping experience",
        description="Usage purpose title",
        required=True
    )

    # Data agreement usage purpose description
    usage_purpose_description = fields.Str(
        data_key="purpose_description",
        example="Collecting user data for offering custom tailored shopping experience",
        description="Usage purpose description",
        required=True
    )

    # Data agreement legal basis
    legal_basis = fields.Str(
        data_key="lawful_basis",
        example="consent",
        description="Legal basis of processing",
        required=True,
        validate=validate.OneOf(
            [
                "consent",
                "legal_obligation",
                "contract",
                "vital_interest",
                "public_task",
                "legitimate_interest",

            ]
        )
    )

    # Data agreement method of use (i.e. how the data is used)
    # 2 method of use: "data-source" and "data-using-service"
    method_of_use = fields.Str(
        data_key="method_of_use",
        example="data-using-service",
        description="Method of use (or data exchange mode)",
        required=True,
        validate=validate.OneOf(
            [
                "data-source",
                "data-using-service",

            ]
        )
    )

    # Data agreement data policy
    data_policy = fields.Nested(
        DataAgreementDataPolicySchema,
        required=True,
        description="Data policy"
    )

    # Data agreement personal data (attributes)
    personal_data = fields.List(
        fields.Nested(DataAgreementPersonalDataSchema),
        required=True,
        description="Personal data (attributes)"
    )

    # Data agreement DPIA metadata
    dpia = fields.Nested(
        DataAgreementDPIASchema,
        description="DPIA metadata"
    )

    # Data agreement events
    event =fields.List(fields.Nested(DataAgreementEventSchema))

    # Data agreement proof
    proof = fields.Nested(DataAgreementProofSchema)

    # Data agreement principle did
    principle_did = fields.Str(
        data_key="data_subject_did",
        example="did:mydata:123456789abcdefghi",
        description="Principle did"
    )


import re

from marshmallow.validate import Regexp

MYDATA_DID_REGEX = "did:mydata(:?(?P<did_type>0|1|2|3|4))?:(?P<identifier>z[a-km-zA-HJ-NP-Z1-9]+)"
MYDATA_DID_PATTERN = re.compile(f"^{MYDATA_DID_REGEX}$")

class MyDataDID(Regexp):
    """Validate value against MyData DID."""

    EXAMPLE = "z6MkfiSdYhnLnS6jfwSf2yS2CiwwjZGmFUFL5QbyL2Xu8z2E"
    PATTERN = rf"^did:mydata(:?(?P<did_type>0|1|2|3|4))?:(?P<identifier>z[a-km-zA-HJ-NP-Z1-9]+)"

    def __init__(self):
        """Initializer."""

        super().__init__(
            MyDataDID.PATTERN,
            error="Value {input} is not an mydata decentralized identifier (DID)",
        )

MYDATA_DID = {"validate": MyDataDID(), "example": MyDataDID.EXAMPLE}


if __name__ == "__main__":
    print(MYDATA_DID_PATTERN.match("did:mydata:0:z6MkfiSdYhnLnS6jfwSf2yS2CiwwjZGmFUFL5QbyL2Xu8z2E").group('did_type'))
    print(MYDATA_DID_PATTERN.match("did:mydata:z6MkfiSdYhnLnS6jfwSf2yS2CiwwjZGmFUFL5QbyL2Xu8z2E").group('did_type'))


from marshmallow import fields, EXCLUDE
from aries_cloudagent.messaging.models.base import BaseModel, BaseModelSchema


class DataAgreementNegotiationReceiptBody(BaseModel):
    """Data Agreement Negotiation Receipt Body"""

    class Meta:
        schema_class = "DataAgreementNegotiationReceiptBodySchema"

    def __init__(
        self,
        *,
        instance_id: str,
        blockchain_receipt: dict,
        blink: str,
        mydata_did: str,
        **kwargs
    ):
        super().__init__(**kwargs)

        # Set model attributes.
        self.blockchain_receipt = blockchain_receipt
        self.blink = blink
        self.mydata_did = mydata_did
        self.instance_id = instance_id


class DataAgreementNegotiationReceiptBodySchema(BaseModelSchema):
    """Data Agreement Negotiation Receipt Body schema"""

    class Meta:
        model_class = DataAgreementNegotiationReceiptBody

        # Unknown fields are excluded
        unknown = EXCLUDE

    instance_id = fields.Str()
    blockchain_receipt = fields.Str()
    blink = fields.Str()
    mydata_did = fields.Str()



from aries_cloudagent.messaging.agent_message import AgentMessage, AgentMessageSchema
from marshmallow import EXCLUDE, fields
from ..message_types import EXISTING_CONNECTIONS, PROTOCOL_PACKAGE
from ..models.existing_connections_model import (
    ExistingConnectionsBody,
    ExistingConnectionsBodySchema
)

# Handler class for existing connections message.
HANDLER_CLASS = (
    f"{PROTOCOL_PACKAGE}.handlers"
    ".existing_connections_handler.ExistingConnectionsMessageHandler"
)


class ExistingConnectionsMessage(AgentMessage):
    """
    Message class for existing connections.
    """

    class Meta:

        # Handler class that can handle this message
        handler_class = HANDLER_CLASS

        # Message type
        message_type = EXISTING_CONNECTIONS

        # Message schema class
        schema_class = "ExistingConnectionsMessageSchema"

    def __init__(
        self,
        *,
        body: ExistingConnectionsBody,
        **kwargs
    ):
        """
        Initialize a ExistingConnectionsMessage message instance.
        """
        super().__init__(**kwargs)

        # Message body
        self.body = body


class ExistingConnectionsMessageSchema(AgentMessageSchema):
    """
    Schema class for existing connections message.
    """

    class Meta:
        # The message class that this schema is for
        model_class = ExistingConnectionsMessage

        # Unknown fields to exclude from the schema (handled by marshmallow)
        unknown = EXCLUDE

    # Message body
    body = fields.Nested(
        ExistingConnectionsBodySchema,
        required=True
    )



from marshmallow import fields
from aries_cloudagent.messaging.models.base import BaseModel, BaseModelSchema
from aries_cloudagent.messaging.valid import UUIDFour

from .data_agreement_negotiation_offer_model import (
    DataAgreementProof,
    DataAgreementProofSchema,
    DataAgreementEvent,
    DataAgreementEventSchema
)


class DataAgreementNegotiationRejectBody(BaseModel):
    """Data Agreement Negotiation Reject Body"""

    class Meta:
        """Data Agreement Negotiation Reject Body metadata"""

        schema_class = "DataAgreementNegotiationRejectBodySchema"

    def __init__(
        self,
        *,
        data_agreement_id: str = None,
        event: DataAgreementEvent = None,
        proof: DataAgreementProof = None,
        **kwargs
    ):
        """Data Agreement Negotiation Reject Body init"""

        super().__init__(**kwargs)

        # Set attributes
        self.data_agreement_id = data_agreement_id
        self.event = event
        self.proof = proof


class DataAgreementNegotiationRejectBodySchema(BaseModelSchema):
    """Data Agreement Negotiation Reject Body schema"""

    class Meta:
        """Data Agreement Negotiation Reject Body schema metadata"""

        model_class = DataAgreementNegotiationRejectBody

    # Data agreement id
    data_agreement_id = fields.Str(
        data_key="id",
        example=UUIDFour.EXAMPLE,
        description="Data agreement identifier"
    )

    # Data agreement events
    event = fields.Nested(DataAgreementEventSchema)

    # Data agreement proof
    proof = fields.Nested(DataAgreementProofSchema)



from ...message_types import SPEC_URI


TAGS_MYDATA_DID_OPERATIONS_LABEL = "Data Agreement - MyData DID Operations"
TAGS_MYDATA_DID_OPERATIONS = {
    "name": TAGS_MYDATA_DID_OPERATIONS_LABEL,
    "description": "MyData DID Protocol 1.0 (ADA RFC 0001)",
    "externalDocs": {
        "description": "Specification",
        "url": SPEC_URI,
    },
}

TAGS_DATA_AGREEMENT_CORE_FUNCTIONS_LABEL = "Data Agreement - Core Functions"
TAGS_DATA_AGREEMENT_CORE_FUNCTIONS = {
    "name": TAGS_DATA_AGREEMENT_CORE_FUNCTIONS_LABEL,
    "description": "Data Agreement Protocol 1.0 (ADA RFC 0002)",
    "externalDocs": {
        "description": "Specification",
        "url": SPEC_URI,
    },
}


TAGS_DATA_AGREEMENT_AUDITOR_FUNCTIONS_LABEL = "Data Agreement - Auditor Functions"
TAGS_DATA_AGREEMENT_AUDITOR_FUNCTIONS = {
    "name": TAGS_DATA_AGREEMENT_AUDITOR_FUNCTIONS_LABEL,
    "description": "Data Agreement Proofs Protocol 1.0 (ADA RFC 0004)",
    "externalDocs": {
        "description": "Specification",
        "url": SPEC_URI,
    },
}


TAGS_JSONLD_FUNCTIONS_LABEL = "JSON-LD"
TAGS_JSONLD_FUNCTIONS = {
    "name": TAGS_JSONLD_FUNCTIONS_LABEL,
    "description": "JSON-LD functions",
}


TAGS_DATA_CONTROLLER_FUNCTIONS_LABEL = "Data Controller"
TAGS_DATA_CONTROLLER_FUNCTIONS = {
    "name": TAGS_DATA_CONTROLLER_FUNCTIONS_LABEL,
    "description": "Data Controller functions",
}



import json
import logging
from aiohttp import web
from aiohttp_apispec import (
    docs,
    querystring_schema,
    response_schema,
    match_info_schema,
)

from aries_cloudagent.messaging.models.base import BaseModelError
from aries_cloudagent.storage.error import StorageError
from aries_cloudagent.storage.base import BaseStorage
from aries_cloudagent.storage.indy import IndyStorage
from aries_cloudagent.storage.record import StorageRecord

from ..manager import ADAManager, ADAManagerError

from ..routes.maps.tag_maps import (
    TAGS_MYDATA_DID_OPERATIONS_LABEL
)

from .openapi import (
    SendReadDIDMessageMatchInfoSchema,
    MyDataDIDRemoteRecordsQueryStringSchema,
    MyDataDIDRemoteRecordResponseSchema
)

LOGGER = logging.getLogger(__name__)

PAGINATION_PAGE_SIZE = 10


@docs(
    tags=[TAGS_MYDATA_DID_OPERATIONS_LABEL],
    summary="Send read-did didcomm message to MyData DID registry",
)
@match_info_schema(SendReadDIDMessageMatchInfoSchema())
async def send_read_did_message_to_mydata_did_registry(request: web.BaseRequest):
    """
    Request handler for sending read-did didcomm message to MyData DID registry
    """

    # Context
    context = request.app["request_context"]

    # did:mydata identifier
    did = request.match_info["did"]

    result = {}

    try:

        # Initialize MyData DID manager
        mydata_did_manager = ADAManager(context=context)

        # Send read-did message to MyData DID registry
        transaction_record = await mydata_did_manager.send_read_did_message(did=did)

        if transaction_record:
            temp_record = transaction_record.serialize()
            temp_messages_list = []
            for message in temp_record.get("messages_list", []):
                temp_messages_list.append(json.loads(message))

                temp_record["messages_list"] = temp_messages_list

            result = temp_record

        else:
            raise web.HTTPInternalServerError(reason="Failed to send read-did message")

    except ADAManagerError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response(result)


@docs(
    tags=[TAGS_MYDATA_DID_OPERATIONS_LABEL],
    summary="Fetch MyData DID remote records.",
)
@querystring_schema(MyDataDIDRemoteRecordsQueryStringSchema())
@response_schema(MyDataDIDRemoteRecordResponseSchema(many=True), 200)
async def mydata_did_remote_records_list(request: web.BaseRequest):
    """
    Request handler for fetching MyData DID remote records
    """

    # Context
    context = request.app["request_context"]

    # Query string parameters
    tag_filter = {}
    if "did" in request.query and request.query["did"] != "":
        tag_filter["did"] = request.query["did"]

    if "sov_verkey" in request.query and request.query["sov_verkey"] != "":
        tag_filter["sov_verkey"] = request.query["sov_verkey"]

    if "status" in request.query and request.query["status"] != "":
        tag_filter["status"] = request.query["status"]

    results = []

    try:
        # Storage
        storage: IndyStorage = await context.inject(BaseStorage)

        # Search remote records
        remote_records: StorageRecord = await storage.search_records(
            type_filter=ADAManager.RECORD_TYPE_MYDATA_DID_REMOTE, tag_query=tag_filter
        ).fetch_all()

        for remote_record in remote_records:
            results.append(
                {
                    "did": remote_record.tags.get("did"),
                    "sov_verkey": remote_record.tags.get("sov_verkey"),
                    "status": remote_record.tags.get("status"),
                    "diddoc": json.loads(remote_record.value),
                }
            )

    except (StorageError, BaseModelError) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response(results)



from marshmallow import EXCLUDE

from aries_cloudagent.messaging.ack.message import Ack, AckSchema

from ..message_types import DATA_AGREEMENT_TERMINATION_TERMINATE_ACK, PROTOCOL_PACKAGE

HANDLER_CLASS = (
    f"{PROTOCOL_PACKAGE}.handlers.data_agreement_termination_ack_handler.DataAgreementTerminationAckHandler"
)


class DataAgreementTerminationAck(Ack):
    """Base class representing an explicit ack message for data agreement termination protocol."""

    class Meta:
        """DataAgreementTerminationAck metadata."""

        handler_class = HANDLER_CLASS
        message_type = DATA_AGREEMENT_TERMINATION_TERMINATE_ACK
        schema_class = "DataAgreementTerminationAckSchema"

    def __init__(self, status: str = None, **kwargs):
        """
        Initialize an explicit ack message instance.

        Args:
            status: Status (default OK)

        """
        super().__init__(status, **kwargs)


class DataAgreementTerminationAckSchema(AckSchema):
    """Schema for DataAgreementTerminationAck class."""

    class Meta:
        """DataAgreementTerminationAck schema metadata."""

        model_class = DataAgreementTerminationAck
        unknown = EXCLUDE



from aries_cloudagent.messaging.agent_message import AgentMessage, AgentMessageSchema

from marshmallow import EXCLUDE, fields

from ..message_types import PROTOCOL_PACKAGE, READ_DID_RESPONSE
from ..models.diddoc_model import MyDataDIDResponseBody, MyDataDIDResponseBodySchema
from ..utils.regex import MYDATA_DID


# Handler class for /mydata/v1.0/read-did-response
HANDLER_CLASS = (
    f"{PROTOCOL_PACKAGE}.handlers"
    ".read_did_response_handler.ReadDIDResponseHandler"
)

class ReadDIDResponseMessage(AgentMessage):
    """
    Message class for response to reading a DID.
    """
    class Meta:

        # Handler class that can handle this message
        handler_class = HANDLER_CLASS

        # Message type
        message_type = READ_DID_RESPONSE

        # Message schema class
        schema_class = "ReadDIDResponseMessageSchema"

    def __init__(self, *, from_did, to_did, created_time, body: MyDataDIDResponseBody, **kwargs):
        """
        Initialize a ReadDIDResponseMessage message instance.
        """
        super().__init__(**kwargs)

        # Set attributes
        self.from_did = from_did
        self.to_did = to_did
        self.created_time = created_time
        self.body = body

class ReadDIDResponseMessageSchema(AgentMessageSchema):
    """
    Schema class for response to reading a DID.
    """
    class Meta:
        # Model class
        model_class = ReadDIDResponseMessage

        # Unknown fields to exclude from the schema (handled by marshmallow)
        unknown = EXCLUDE
    
    # From DID
    from_did = fields.Str(data_key="from", **MYDATA_DID)

    # To DID
    to_did = fields.Str(data_key="to", **MYDATA_DID)

    # Created Time
    created_time = fields.Str(data_key="created_time")

    # Message body
    body = fields.Nested(MyDataDIDResponseBodySchema, required=True)
